#!/usr/bin/python

"""
Usage:
    # Interface configuration
    iface (add|del) bond <interface>
    iface (add|del) bridge
    iface (add|del) loopback lo
    iface (add|del) vrf <interface>

    iface (add|del) interface <interface>
    iface add vlan-interface <interface> [vlan <0-4096>]
    iface del vlan-interface <interface>
    iface add vni <interface> [vlan <0-4096>]
    iface del vni <interface>
    iface (add|del) interface <interface> bridge-trunk [vlans <number-range-list>]
    iface (add|del) vlan <number-range-list>
Options:
    <ipv4>                       : An IPv4 address
    <ipv6>                       : An IPv6 address
    <ip>                         : An IPv4 or IPv6 Address
    <ipv4/prefixlen>             : An IPv4 address and prefix length
    <ipv6/prefixlen>             : An IPv6 address and prefix length
    <ip/prefixlen>               : An IPv4 or IPv6 address and prefix length
    <mac>                        : A MAC address
    <text>                       : Any text
    <number>                     : Any integer
    <interface>                  : An interface name "swp1" or glob "swp1-4,6,10-12"
    <ipv4-vrf-text>              : Example: "<ipv4> vrf <text>"
    <number-ipv4-list>           : Example: "100=10.0.100.1,101=10.101.1"
    <interface-list>             : Example: "swp1,swp2,swp3"
    <number-range-list>          : Example: "1-5,8,10-12,15"
    <interface-range-list>       : Example: "swp1=100 swp2=100"
    <mac-ip/prefixlen-list>      : Example: "00:11:22:33:44:01,10.0.0.1/24,10.0.1.2/24"
    <number-interface-list>      : Example: "4,swp1,swp2"
    <interface-yes-no-list>      : Example: "swp1=yes,swp2=no"
    <interface-yes-no-0-1-list>  : Example: "swp1=yes,swp2=no,swp3=1,swp4=0"
    <interface-yes-no-auto-list> : Example: "swp1=yes,swp2=no,swp3=auto"
    bond                         : bond, port-channel, etc
    bridge                       : A layer2 bridge
    vlan                         : Virtual Local Area Network
    vlan-interface               : Virtual Local Area Network interface
    vni                          : VXLAN Network Identifier interface
    version                      : version number
    configuration                : configuration
    hostname                     : system hostname
    verbose                      : Show detailed output
    detail                       : Show detailed output
    details                      : Show detailed output
    vxlan                        : Virtual Extensible LAN
    bridge-trunk                 : A layer2 trunk, bridge-trunk is an alias for bridge-vids
    100                          : 100M
    1000                         : 1G
    10000                        : 10G
    40000                        : 40G
    100000                       : 100G
    bond-slaves                  : Add physical ports to a bond
    no                           : no means no
    yes                          : yes means yes
"""

# At one point we experimented with a bond-master keyword but bond that
# bonds are pretty configurable already
#
#    iface (add|del) interface <interface> bond-master <interface-bond>
#    bond-master                  : Add a physical interface to a bond

from collections import OrderedDict
from ipaddr import IPv6Network, IPv4Network, IPv4Address, IPv6Address
from network_docopt import sort_ifaces, sort_for_humans
from nclu import (print_diff,
    ifname_is_valid,
    ifname_is_globish,
    ifname_is_glob,
    ifname_expand_glob,
    is_number_glob,
    glob_to_numbers,
    numbers_to_glob,
    print_pending_add_del_commands)
from pprint import pformat
from shutil import rmtree, copyfile
from subprocess import check_output, STDOUT, CalledProcessError
import logging
import logging.handlers
import os
import re
import sys
sys.path.append('/usr/share/ifupdown2/addons/')

log = logging.getLogger('netd')
addons_cmds = []

all_interface_keywords = (
    'address',
    'post-down',
    'post-up',
    'pre-down',
    'pre-up')

bond_interface_keywords = (
    'address-virtual',
    'alias',
    'bond-lacp-bypass-allow',
    'bond-lacp-rate',
    'bond-miimon',
    'bond-min-links',
    'bond-slaves',
    'bridge-access',
    'bridge-vids',
    'clag-id',
    'clagd-args',
    'clagd-backup-ip',
    'clagd-enable',
    'clagd-peer-ip',
    'clagd-priority',
    'clagd-sys-mac',
    'gateway',
    'mstpctl-bpduguard',
    'mstpctl-portadminedge',
    'mstpctl-portautoedge',
    'mstpctl-portbpdufilter',
    'mstpctl-portnetwork',
    'mstpctl-treeprio',
    'mtu',
    'vrf')

bridge_interface_keywords = (
    'address-virtual',
    'alias',
    'bridge-ageing',
    'bridge-allow-untagged',
    'bridge-igmp-querier-src',
    'bridge-mcqifaddr',
    'bridge-mcquerier',
    'bridge-mcsnoop',
    'bridge-ports',
    'bridge-pvid',
    'bridge-stp',
    'bridge-vids',
    'bridge-vlan-aware',
    'gateway',
    'mtu')

loopback_interface_keywords = (
    'alias',
    'clagd-vxlan-anycast-ip',
    'mtu',
    'vxrd-src-ip',
    'vxrd-svcnode-ip',
    'vrf')

physical_interface_keywords = (
    'address-virtual',
    'alias',
    'bridge-access',
    'bridge-vids',
    'clag-id',
    'clagd-args',
    'clagd-backup-ip',
    'clagd-enable',
    'clagd-peer-ip',
    'clagd-priority',
    'clagd-sys-mac',
    'gateway',
    'link-autoneg',
    'link-duplex',
    'link-speed',
    'mstpctl-bpduguard',
    'mstpctl-portadminedge',
    'mstpctl-portautoedge',
    'mstpctl-portbpdufilter',
    'mstpctl-portnetwork',
    'mstpctl-treeprio',
    'mtu',
    'vrf')

vlan_interface_keywords = (
    'address-virtual',
    'alias',
    'gateway',
    'mtu',
    'vlan-id',
    'vlan-raw-device',
    'vrf')

vni_interface_keywords = (
    'alias',
    'mtu',
    'vxlan-id',
    'vxlan-local-tunnelip',
    'vxlan-remoteip',
    'vxlan-svcnodeip')

vrf_interface_keywords = ('vrf-table',)

all_keywords = list(set((all_interface_keywords +
    bond_interface_keywords +
    bridge_interface_keywords +
    loopback_interface_keywords +
    physical_interface_keywords +
    vlan_interface_keywords +
    vni_interface_keywords +
    vrf_interface_keywords)))


def ifname_in_etc_network_interfaces(ifname):

    # If this file exists it means the user is in the process of
    # adding ifname
    if os.path.exists('/var/run/nclu/iface/%s' % ifname):
        return True
    else:
        iface_line = 'iface %s' % ifname

        with open('/etc/network/interfaces', 'r') as fh:
            for line in fh.readlines():
                if line.strip() == iface_line:
                    return True
    return False


def standardize_indent(line):
    """
    /etc/network/interfaces tends to be indented with four spaces but ifquery uses tabs
    """
    indented = line.startswith(' ') or line.startswith('\t')
    line = line.strip()

    if line and indented:
        return '    ' + line
    else:
        return line


def write_new_interface(lines_to_write, prev_line, ifname, lines, persist):

    # Always have a blank line between interface stanzas
    if prev_line:
        lines_to_write.append("\n")

    for line in lines:
        lines_to_write.append(line)
    lines_to_write.append("\n")

    if persist:
        os.remove("/var/run/nclu/iface/%s" % ifname)


class Command(object):
    """
    Base class for the various CommandXYZ classes that are used to represent
    commands supported by ifupdown2 addon modules
    """

    def __init__(self, keyword, value):
        """
        value is a dictionary of attributes for the 'keyword' command.
        value is extracted from the _modinfo dictionary for the ifupdown2
        addon that owns this command.
        """
        self.keyword = keyword
        self.validvals = None
        self.rangemin = None
        self.rangemax = None
        self.deprecated = False
        self.help = value.get('help')
        self.usage_str_add = []
        self.usage_str_del = []
        self.validvals = []
        add_string = None
        del_string = None

        # There is some weird ifupdown1 historical reason that the address.py
        # addon cannot list <ipv4> as the validvals for netmask so we hard code
        # it here to work around that
        if self.keyword == 'netmask' and 'validvals' not in value:
            value['validvals'] = ['<ipv4>', ]

        if 'deprecated' in value:
            self.deprecated = value['deprecated']

        if 'validvals' in value:
            sorted_validvals = sorted(value['validvals'])

            # Compress the various ways of saying "yes or no" down to "yes or no"
            if (sorted_validvals == ['no', 'off', 'on', 'yes'] or
                sorted_validvals == ['0', '1', 'no', 'yes'] or
                sorted_validvals == ['no', 'yes']):
                value['validvals'] = ['yes', 'no']

            # bond-mode takes a number of keyword, chop the numbers
            elif sorted_validvals == ['0', '1', '2', '3', '4', '5', '6', '802.3ad', 'active-backup', 'balance-alb', 'balance-rr', 'balance-tlb', 'balance-xor', 'broadcast']:
                value['validvals'] = ['802.3ad', 'active-backup', 'balance-alb', 'balance-rr', 'balance-tlb', 'balance-xor', 'broadcast']


            # These are keywords that can be configured under the bridge via <interface-yes-no-list>
            # or under the swp via 'yes/no'. We will restrict these keywords to just the swp so change
            # the validvals to 'yes/no'
            if self.keyword in ('mstpctl-bpduguard',
                                'mstpctl-portadminedge',
                                'mstpctl-portautoedge',
                                'mstpctl-portbpdufilter',
                                'mstpctl-portnetwork',
                                'mstpctl-treeprio'):
                value['validvals'] = ['yes', 'no']

            for val in value['validvals']:

                if isinstance(val, str):
                    if val == '<ipaddr>':
                        self.validvals.append('<ip>')

                    elif val == '<ipaddr/prefixlen>':
                        self.validvals.append('<ip/prefixlen>')

                    elif val == '<interface>':
                        self.validvals.append('<interface-as-arg>')

                    # 1.1.1.1 or "1.1.1.1 vrf BLUE"
                    elif val == '<ipv4-vrf-text>':
                        self.validvals.append('<ipv4> [vrf <text>]')

                    else:
                        self.validvals.append(val)

                elif isinstance(val, int):
                    self.validvals.append(str(val))

                else:
                    raise Exception("Unexpected validvals type %s for %s" % (type(val), val))

            add_string = "    iface add interface <interface> %s " % self.keyword
            del_string = "    iface del interface <interface> %s " % self.keyword

            if len(self.validvals) == 1:
                add_string += self.validvals[0]
                del_string += '[' + self.validvals[0] + ']'
            else:
                # (no|yes) just looks weird so do not sort this one
                if self.validvals == ['yes', 'no']:
                    add_string += "(yes|no)"
                    del_string += "[yes|no]"
                else:
                    add_string += "(%s)" % '|'.join(sort_for_humans(self.validvals))
                    del_string += "[%s]" % '|'.join(sort_for_humans(self.validvals))

        elif 'validrange' in value:
            self.rangemin = int(value['validrange'][0])
            self.rangemax = int(value['validrange'][1])
            add_string = "    iface add interface <interface> %s <%d-%d>" % (self.keyword, self.rangemin, self.rangemax)
            del_string = "    iface del interface <interface> %s [<%d-%d>]" % (self.keyword, self.rangemin, self.rangemax)

        else:
            add_string = "    iface add interface <interface> %s <wildcard>" % self.keyword
            del_string = "    iface del interface <interface> %s [<wildcard>]" % self.keyword

        # This is a special case...we override 'address (<ipv4/prefixlen>|<ipv6/prefixlen>)' with
        # "[ip] address (<ipv4/prefixlen>|<ipv6/prefixlen>|dhcp)" command
        if self.keyword == 'address':
            add_string = "    iface add interface <interface> [ip] address (<ipv4/prefixlen>|<ipv6/prefixlen>|dhcp)"
            del_string = "    iface del interface <interface> [ip] address [<ipv4/prefixlen>|<ipv6/prefixlen>|dhcp]"
            self.validvals.append('dhcp')

        keyword_scope_defined = False

        int_types = ('bond', 'bridge', 'loopback', 'vlan-interface', 'vni', 'vrf')
        if self.keyword in all_interface_keywords:
            keyword_scope_defined = True
            self.usage_str_add.append(add_string)
            self.usage_str_del.append(del_string)

            for int_type in int_types:

                # bridge type is special, we restrict its name to 'bridge'
                if int_type == 'bridge':
                    tmp_add_string = add_string.replace(' interface <interface> ', ' bridge ')
                    tmp_del_string = del_string.replace(' interface <interface> ', ' bridge ')

                elif int_type == 'loopback':
                    tmp_add_string = add_string.replace(' interface <interface> ', ' loopback lo ')
                    tmp_del_string = del_string.replace(' interface <interface> ', ' loopback lo ')

                else:
                    tmp_add_string = add_string.replace(' interface ', ' %s ' % int_type)
                    tmp_del_string = del_string.replace(' interface ', ' %s ' % int_type)

                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

        else:
            if self.keyword in bridge_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface <interface> ', ' bridge ')
                tmp_del_string = del_string.replace(' interface <interface> ', ' bridge ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword in bond_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' bond ')
                tmp_del_string = del_string.replace(' interface ', ' bond ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword in loopback_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface <interface> ', ' loopback lo ')
                tmp_del_string = del_string.replace(' interface <interface> ', ' loopback lo ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword in physical_interface_keywords:
                keyword_scope_defined = True
                self.usage_str_add.append(add_string)
                self.usage_str_del.append(del_string)

            if self.keyword in vlan_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' vlan-interface ')
                tmp_del_string = del_string.replace(' interface ', ' vlan-interface ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword in vni_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' vni ')
                tmp_del_string = del_string.replace(' interface ', ' vni ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword in vrf_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' vrf ')
                tmp_del_string = del_string.replace(' interface ', ' vrf ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

        if not keyword_scope_defined:
            # We currently do not do anything with the deprecated commands
            # so don't bother spewing the log for these
            if not self.deprecated:
                log.info('Where does keyword %s belong?' % self.keyword)
            self.usage_str_add.append(add_string)
            self.usage_str_del.append(del_string)

    def value_from_args(self, args):
        args_value_keyword = None

        if self.validvals:
            for val in self.validvals:
                # Combos such as '<ipv4> [vrf <text>]' are a special case
                if ' ' in val:
                    values = []

                    for val_component in val.split():
                        if val_component.startswith('['):
                            val_component = val_component[1:]

                        if val_component.endswith(']'):
                            val_component = val_component[:-1]

                        tmp = args.get(val_component)
                        if tmp:
                            values.append(tmp)

                    if values:
                        return ' '.join(values)
                else:
                    if args.get(val):
                        args_value_keyword = val
                        break

        elif self.rangemax:
            args_value_keyword = "<%d-%d>" % (self.rangemin, self.rangemax)

        else:
            for x in ('<interface>', '<wildcard>'):
                if args.get(x):
                    args_value_keyword = x

        if args_value_keyword is None:
            return None

        if args_value_keyword.startswith('<'):
            return args.get(args_value_keyword)
        else:
            return args_value_keyword


class CommandAddDelMultiLine(Command):
    """
    Commands that are allowed to appear multiple times in and interface stanza...example:

    auto swp1
    iface swp1
        address 1.1.1.1/24
        address 2.2.2.1/24
    """
    pass


class CommandAddDelMultiValue(Command):
    """
    Commands that can only appear once in an interface stanza but can have a list of values...example:

    auto br0
    iface br0
        bridge-ports swp1 swp2 swp7
    """
    pass


class CommandSetClear(Command):
    """
    Commands that can only appear one time in an interface stanza...example:

    auto vni100
    iface vni100
        vxlan-id 1.1.1.1
    """
    pass


def build_doc_string(blacklist):
    """
    Parse the _modinfo dictionary for the various ifupdownaddons modules.
     _modinfo contains information about the commands supported by the addon.

    We use this to:
    - modify __doc__ to include the CLI to support each command
    - create a Command object for each command
    """
    global __doc__
    global addons_cmds

    state = 'LIMBO'
    usage = []
    options = []
    items = []
    addon_names = []

    # Parse /etc/network/ifupdown2/addons.conf to build a list of all installed addons
    with open('/etc/network/ifupdown2/addons.conf', 'r') as fh:
        for line in fh.readlines():
            line = line.strip()

            if not line:
                continue

            if ',' in line:
                addon = line.split(',')[1].strip()
                addon_names.append(addon)
            else:
                log.info("Line without comma in /etc/network/ifupdown2/addons.conf '%s'" % line)

    addon_names = sorted(list(set(addon_names)))

    for addon in addon_names:
        '''
        It is possible to use the __import__ function to import using a variable.

        package = "os"
        name = "path"

        imported = getattr(__import__(package, fromlist=[name]), name)
        is equivalent to

        from os import path as imported
        '''
        try:
            addon_module = getattr(__import__(addon, fromlist=[addon]), addon)
            log.info("iface imported %s" % addon)
        except ImportError as e:
            log.info("iface failed to import %s due to\n%s\n" % (addon, e))
            continue

        try:
            items += addon_module._modinfo['attrs'].items()

        # There might be an addon that does not have a _modinfo
        except AttributeError:
            pass

    addons = OrderedDict(sorted(items))

    # Parse all of the addons 'attrs' dictionaries to build a list of Command objects
    for (keyword, value) in addons.iteritems():

        # Some keywords are corner cases so we do not list them in order to
        # avoid clutter when the user does TAB complete
        if keyword in blacklist:
            continue

        # Commands like 'address' where you are allowed to have more than one of
        # them per interface
        if 'multiline' in value:
            cmd = CommandAddDelMultiLine(keyword, value)

        # Commands like "bridge-ports" where you can have
        # "bridge-ports swp2 swp3 swp4"
        elif 'multivalue' in value:
            cmd = CommandAddDelMultiValue(keyword, value)

        # Commands where you can only have one (vxlan-id)
        else:
            cmd = CommandSetClear(keyword, value)
        addons_cmds.append(cmd)

    # ====================
    # __doc__ construction
    # ====================

    # Parse the initial docstring, extract the Usage and Options lines
    for line in __doc__.splitlines():
        if line == 'Usage:':
            state = 'USAGE'
        elif line == 'Options:':
            state = 'OPTIONS'
        else:
            if state == 'LIMBO':
                pass
            elif state == 'USAGE':
                usage.append(line)
            elif state == 'OPTIONS':
                options.append(line)
            else:
                raise Exception("Invalid state %s, line is %s" % (state, line))

    addon_usage = []
    # Build all of the Usage and Options lines
    for cmd in addons_cmds:
        if not cmd.deprecated:
            for line in cmd.usage_str_add:
                addon_usage.append(line)

            for line in cmd.usage_str_del:
                addon_usage.append(line)

            if cmd.help:
                options.append("%s : %s" % (cmd.keyword, cmd.help))

    addon_usage.sort()
    usage.extend(addon_usage)

    __doc__ = '\nUsage:\n' + '\n'.join(usage) + '\n\nOptions:\n' + '\n'.join(options)

    return __doc__
    # End of __doc__ construction


class EtcNetworkInterface(object):
    """
    Used to represent an interface, both its state in /etc/network/interfaces
    and in its iface file in /var/run/iface
    """
    def __init__(self, name, ifquery_all):
        self.content = []
        self.name = str(name)
        self.filename = "/var/run/nclu/iface/%s" % self.name
        self.ifquery = []

        # Extract the config for this interface from the 'ifquery -a' output
        found_name = False

        for line in ifquery_all.splitlines():
            line = standardize_indent(line)
            re_name = re.search("^auto\s+%s$" % self.name, line)

            if re_name:
                found_name = True
                self.ifquery.append(line)

            elif found_name:
                # we've started parsing the next interface
                if line.startswith('auto ') or line.startswith('source '):
                    break
                else:
                    if line:
                        self.ifquery.append(line)

        if os.path.isfile(self.filename):
            self.load_iface_state()
        else:
            self.init_iface_state()

    def __str__(self):
        return self.name

    def add_cmd(self, cmd):
        self.content.append(cmd)

    def del_cmd(self, cmd, match_exact, give_feedback):

        # If an interface is configured with "vxlan-id 500" and the user
        # does "net del int swp1 vxlan-id 500" then look for this exact line
        if match_exact:
            if cmd not in self.content:
                if give_feedback:
                    print "%s's configuration does not have '%s'" % (self, cmd.strip())
            else:
                self.content.remove(cmd)

        # But if the user did "net del int swp1 vxlan-id" then we just need to
        # look for the lines that start with "    vxlan-id "
        else:
            lines_to_remove = []

            # Do not change the size of the list as we are looping over it, store
            # the ones to delete and then come back and delete them
            for tmp_cmd in self.content:
                if tmp_cmd.startswith('%s ' % cmd):
                    lines_to_remove.append(tmp_cmd)

            if lines_to_remove:
                for tmp_cmd in lines_to_remove:
                    self.content.remove(tmp_cmd)
            else:
                if give_feedback:
                    print "%s's configuration does not have '%s'" % (self, cmd.strip())

    def mod_cmd(self, old_cmd, new_cmd, give_feedback):

        if old_cmd == new_cmd:
            if give_feedback:
                print "%s's configuration already has '%s'" % (self, new_cmd.strip())
            return

        for (i, line) in enumerate(self.content):
            if line == old_cmd:
                self.content[i] = new_cmd
                return

    def get_cmd(self, keyword):
        keyword = keyword.strip() + ' '

        for line in self.content:
            if line.lstrip().startswith(keyword):
                return line
        return None

    def has_cmd(self, keyword):
        if self.get_cmd(keyword) is None:
            return False
        return True

    def get_keyword_value(self, keyword):
        # line will be something like "mtu 1500" or "bridge-ports swp1 swp2" so
        # split it and return all values after the keyword
        line = self.get_cmd(keyword)

        if line:
            line = line.strip()

            if ' ' in line:
                return ' '.join(line.split()[1:])
        return None

    def is_bond(self):
        if self.has_cmd('bond-slaves'):
            return True
        return False

    def is_bridge(self):
        if self.has_cmd('bridge-ports') or self.has_cmd('bridge-vlan-aware'):
            return True
        return False

    def is_layer2(self):
        if self.has_cmd('bridge-access') or self.has_cmd('bridge-vids'):
            return True

        # You also need to look at the bridge bridge-ports to see if this iface is listed
        return False

    def is_loopback(self):
        for line in self.content:
            if line.rstrip().endswith('inet loopback'):
                return True
        return False

    def is_vni(self):
        if self.has_cmd('vxlan-id') or self.has_cmd('vxlan-local-tunnelip'):
            return True
        return False

    def is_vlan_interface(self):
        if self.has_cmd('vlan-raw-device') or re.search('^bridge\.\d+$', self.name):
            return True
        return False

    def is_vrf(self):
        if self.has_cmd('vrf-table'):
            return True
        return False

    def exists(self):
        # If we are editing this interface via iface then the interface exists
        if os.path.isfile(self.filename):
            return True

        # self.content will contain the output of ifquery.  If the interface
        # is not defined in /etc/network/interfaces then ifquery will return
        # an error so self.content will be empty.
        else:
            if self.content:
                return True
            return False

    def write_file(self):
        """
        Write self.content to a file in /var/run/nclu/iface/
        """
        with open(self.filename, 'w') as fh:
            commands_to_sort = []
            found_iface = False

            for line in self.content:

                if found_iface:
                    commands_to_sort.append(line)

                # 99% of the time this will only be the 'auto' and 'iface' lines
                else:
                    fh.write("%s\n" % line)

                if line.startswith('iface '):
                    found_iface = True

            # Sort all of the commands in this stanza.  This is
            # to make the stanza easier for the user to read.
            for line in sort_for_humans(commands_to_sort):
                fh.write("%s\n" % line)

    def add_interface(self):
        if self.content:
            print "%s already exists in /etc/network/interfaces, no need to add it" % self.name
            return

        # Add the basic commands needed (auto and iface) for this interface
        self.add_cmd("auto %s" % self.name)
        self.add_cmd("iface %s" % self.name)
        self.write_file()

    def del_interface(self):
        self.content = []
        filename_del = self.filename + '.del'
        filename_original = self.filename + '.original'
        valid_interface = False

        if os.path.isfile(self.filename):
            if os.path.getsize(self.filename):
                valid_interface = True
            os.remove(self.filename)

        # If the .original file is empty and we are deleting the interface then
        # there is no need to create a .del file, just delete the .original
        if os.path.isfile(filename_original):
            if not os.path.getsize(filename_original):
                os.remove(filename_original)

                if not valid_interface:
                    print "%s does not exist for us to delete" % self.name

                return

        if os.path.isfile(filename_del):
            print "%s is already marked to be deleted via the next 'net commit'" % self.name
            return

        # Write a .del file so we know this is one that we are trying to delete
        # This will be deleted via commit_pending()
        with open(filename_del, 'w') as fh:
            fh.write('')

    def conf_key_value_multiple_lines(self, add, key, value, give_feedback=True):
        """
        iface config lines where there can be more than one.  Example:
            address 1.1.1.1/32
            address 1.1.1.2/32
        """
        iface_dhcp = "iface %s inet dhcp" % self.name
        iface_static = "iface %s" % self.name

        if key == 'address' and value == 'dhcp':
            if add:
                if iface_static in self.content:
                    self.mod_cmd(iface_static, iface_dhcp, False)

                    # Delete any static IPs assigned
                    if self.get_cmd('address'):
                        self.del_cmd('    address', False, give_feedback)
            else:
                if iface_dhcp in self.content:
                    self.mod_cmd(iface_dhcp, iface_static, False)

        else:
            # make sure iface line does not contain 'dhcp'
            if key == 'address':
                if iface_dhcp in self.content:
                    self.mod_cmd(iface_dhcp, iface_static, False)

            if add:
                cmd = "    %s %s" % (key, value)

                if cmd in self.content:
                    if give_feedback:
                        print "%s's configuration already has '%s'" % (self, cmd.strip())
                else:
                    self.add_cmd(cmd)
            else:
                if value:
                    cmd = "    %s %s" % (key, value)
                    self.del_cmd(cmd, True, give_feedback)
                else:
                    cmd = "    %s" % key
                    self.del_cmd(cmd, False, give_feedback)

        self.write_file()

    def conf_key_value_multiple_values(self, add, key, values, give_feedback=True):
        """
        Some commands (bridge-ports for example) take a list of values that
        we need to be able to add/del from
        """
        ifaces_to_add = []

        if type(values) != list:
            if type(values) == int:
                values = [str(values), ]
            elif values:
                values = values.replace(',', ' ').split()

        # Return early here if we know that all we need to do is nuke all
        # lines for this keyword
        if not add and not values:
            cmd = "    %s" % key
            self.del_cmd(cmd, False, give_feedback)
            self.write_file()
            return ifaces_to_add

        # numeric globs...used by bridge-vids
        if is_number_glob(values):
            values = glob_to_numbers(values)

            old_cmd = self.get_cmd("    %s " % key)
            if old_cmd:
                old_values = old_cmd.split()[1:]
                old_values = glob_to_numbers(old_values)

                if add:
                    if old_values:
                        final_values = old_values + values
                    else:
                        final_values = values
                else:
                    for x in values:
                        if x in old_values:
                            old_values.remove(x)
                    final_values = old_values

                if final_values:
                    glob = numbers_to_glob(final_values)
                    new_cmd = "    %s %s" % (key, glob.replace(',', ' '))

                    if new_cmd != old_cmd:
                        self.mod_cmd(old_cmd, new_cmd, give_feedback)
                else:
                    self.del_cmd(old_cmd, True, give_feedback)

            else:
                if add:
                    glob = numbers_to_glob(values)
                    cmd = "    %s %s" % (key, glob.replace(',', ' '))
                    self.add_cmd(cmd)

        # interfaces
        else:
            # Expand all globs so we can build one big list of the interfaces
            # fed to us by the user
            values_globs_expanded = []
            for (i, value) in enumerate(values):

                if ifname_is_glob(value):
                    for ifname in ifname_expand_glob(value):
                        values_globs_expanded.append(ifname)
                else:
                    # These two keywords accept interfaces as their argument
                    if key in ('bridge-ports', 'bond-slaves'):
                        if not ifname_is_valid(value) or ifname_is_globish(value):
                            return ifaces_to_add
                    values_globs_expanded.append(value)
            values = sort_for_humans(list(set(values_globs_expanded)))

            old_cmd = self.get_cmd("    %s " % key)

            if old_cmd:
                old_values = old_cmd.split()[1:]
                if add:
                    final_values = sort_for_humans(list(set(old_values + values)))
                else:
                    for x in values:
                        if x in old_values:
                            old_values.remove(x)
                    final_values = sort_for_humans(old_values)

                if final_values:
                    new_cmd = "    %s %s" % (key, ' '.join(final_values))
                    if new_cmd != old_cmd:
                        self.mod_cmd(old_cmd, new_cmd, give_feedback)

                        if key in ('bridge-ports', 'bond-slaves'):
                            ifaces_to_add = final_values
                else:
                    self.del_cmd(old_cmd, True, give_feedback)
            else:
                if add:
                    cmd = "    %s %s" % (key, ' '.join(values))
                    self.add_cmd(cmd)

                    # Automatically add the interfaces specifed in bridge-ports or bond-slaves
                    if key in ('bridge-ports', 'bond-slaves'):
                        ifaces_to_add = values

        self.write_file()

        return ifaces_to_add

    def conf_key_value_solo(self, set_value, key, value, give_feedback=True):
        """
        iface configs lines where there can be only one. Example:
        bridge_fd 1
        """
        key = key.rstrip()

        if set_value:
            new_cmd = "    %s %s" % (key, value)
            old_cmd = self.get_cmd('    %s ' % key)

            if old_cmd:
                self.mod_cmd(old_cmd, new_cmd, give_feedback)
            else:
                self.add_cmd(new_cmd)
        else:
            if value:
                cmd = "    %s %s" % (key, value)
                self.del_cmd(cmd, True, give_feedback)
            else:
                cmd = "    %s" % key
                self.del_cmd(cmd, False, give_feedback)
        self.write_file()

    def load_iface_state(self):
        """
        Open the iface file for this interface and store the lines of configuration
        in that file as in self.content
        """
        self.content = []

        if os.path.exists(self.filename):
            with open(self.filename, 'r') as fh:
                for line in fh.readlines():
                    line = standardize_indent(line)

                    # ignore blank lines
                    if line:
                        self.content.append(line)
        else:
            log.debug("%s: filename %s does not exist" % (self, self.filename))

    def init_iface_state(self):
        """
        If the user has never run iface then self.filename will not exist.
        Populate self.content with the current /etc/network/interfaces config
        for this interface
        """
        self.content = self.ifquery

        # Remember the /etc/network/interfaces state of this interface. We
        # use this later to determine if someone is modifying the file and
        # modifying via iface (not supported).
        #
        # This file will be removed by commit_pending()
        with open(self.filename + '.original', 'w') as fh:
            for line in self.content:
                fh.write("%s\n" % line)


class ConfigWrapper(object):

    def __init__(self, show_linux_command):

        if not os.path.isdir('/var/run/nclu/iface/'):
            os.makedirs('/var/run/nclu/iface/', mode=0600)

        self.ifquery_all = None
        self.etc_network_interfaces_init_mod_time = None
        self.show_linux_command = show_linux_command
        self.ifaces = {}

    def auto_add_interfaces(self, ifnames):
        for ifname in ifnames:
            if not ifname_in_etc_network_interfaces(ifname):
                iface = EtcNetworkInterface(ifname, self.ifquery_all)
                iface.add_interface()

    def abort_pending(self):
        if os.path.exists('/var/run/nclu/iface/'):
            for filename in os.listdir('/var/run/nclu/iface/'):
                full_filename = os.path.join('/var/run/nclu/iface/', filename)

                if os.path.isdir(full_filename):
                    rmtree(full_filename)
                else:
                    os.remove(full_filename)

        self.ifquery_all = None
        self.ifaces = {}
        self.etc_network_interfaces_init_mod_time = None

    def commit_pending(self, _verbose):
        """
        Make iface changes persistent by updating /etc/network/interfaces
        """
        modified = []

        # First sanity check the config via 'ifup --syntax-check'
        self._save_or_show_pending(persist=False, verbose=False)

        if os.path.exists('/var/run/nclu/iface/interfaces.tmp'):
            try:
                output = check_output(['/sbin/ifup',
                                       '--syntax-check',
                                       '--verbose',
                                       '--all',
                                       '--interfaces',
                                       '/var/run/nclu/iface/interfaces.tmp'],
                                       stderr=STDOUT)
                os.remove('/var/run/nclu/iface/interfaces.tmp')
            except Exception as e:
                print "\nERROR: We cannot commit the current pending changes due to the following errors(s)\n" +\
                      "they would introduce in /etc/network/interfaces:\n\n%s\n\n" %\
                    '\n'.join(line for line in e.output.splitlines() if not line.startswith('info'))

                os.remove('/var/run/nclu/iface/interfaces.tmp')
                return modified

        modified = self._save_or_show_pending(persist=True, verbose=_verbose)
        self.ifquery_all = None
        self.ifaces = {}
        self.etc_network_interfaces_init_mod_time = None

        # Return a list of the files modified...iface can only
        # modify /etc/network/interfaces
        if modified:
            return ['/etc/network/interfaces' ,]
        else:
            return []

    def show_pending(self, print_errors=True):
        """
        print_errors will be False when the user does a 'net commit'.  The commit
        will run --syntax-check and we do not want to double print any errors
        thus the print_errors False.
        """
        self._save_or_show_pending(persist=False, verbose=False)
        diff_pending = print_diff('/etc/network/interfaces', '/var/run/nclu/iface/interfaces.tmp')

        if os.path.exists('/var/run/nclu/iface/interfaces.tmp'):

            if print_errors:
                try:
                    output = check_output(['/sbin/ifup',
                                           '--syntax-check',
                                           '--verbose',
                                           '--all',
                                           '--interfaces',
                                           '/var/run/nclu/iface/interfaces.tmp'],
                                           stderr=STDOUT)
                except Exception as e:
                    print "\nERROR: pending changes will introduce the following error(s) "+\
                        "in /etc/network/interfaces\n\n%s\n\n" % \
                        '\n'.join(line for line in e.output.splitlines() if not line.startswith('info'))

            os.remove('/var/run/nclu/iface/interfaces.tmp')

        return diff_pending

    def show_config(self, user_may_edit):
        print "/etc/network/interfaces"
        print "======================="
        cmd = ['/sbin/ifquery', '-ca']

        if self.show_linux_command:
            print "exec '%s'\n" % ' '.join(cmd)

        output = []

        try:
            raw_output = check_output(cmd).splitlines()
        except CalledProcessError as e:
            raw_output = e.output.splitlines()

        for line in raw_output:
            line = line.rstrip()

            if line.endswith('[pass]'):
                line = line.replace('[pass]', '')

            elif line.endswith('[]'):
                line = line.replace('[]', '')

            # TABs are evil
            line = line.replace('\t', '    ').rstrip()
            output.append(line)

        print '\n'.join(output)

    def vlan_add(self, bridge, vlan_glob):
        bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob)

    def vlan_del(self, bridge, vlan_glob):
        if not bridge:
            return

        bridge.conf_key_value_multiple_values(False, 'bridge-vids', vlan_glob)
        vlans_to_del = glob_to_numbers(vlan_glob)

        # remove any access interface in these vlans
        # remove these vlans from the bridge-vids of all trunks
        for ifquery_line in self.ifquery_all.splitlines():
            ifquery_line = standardize_indent(ifquery_line)
            re_name = re.search("^auto\s+(\S+)\s*$", ifquery_line)

            if re_name and re_name.group(1) != 'bridge':
                ifname = re_name.group(1)
                iface = EtcNetworkInterface(ifname, self.ifquery_all)

                # See if this is an SVI named "bridge.10"...if so and if
                # vlan 10 is one of the vlans we are deleting then delete
                # this interface and 'continue'
                re_strict_svi_name = re.search('^bridge\.(\d+)$', ifname)
                if re_strict_svi_name:
                    vlan_id = int(re_strict_svi_name.group(1))

                    if vlan_id in vlans_to_del:
                        iface.del_interface()
                        continue

                for line in iface.content:

                    # If this interface is in a vlan that is being deleted then:
                    # - remove the bridge-access line
                    # - remove this interface from the bridge's bridge-ports
                    if 'bridge-access' in line:
                        bridge_access_vlan = int(line.strip().split()[1])

                        if bridge_access_vlan in vlans_to_del:
                            iface.conf_key_value_solo(False, 'bridge-access', bridge_access_vlan)
                            bridge.conf_key_value_multiple_values(False, 'bridge-ports', ifname)

                    # If this is a trunk remove the vlans from the trunk's bridge-vids
                    elif 'bridge-vids' in line:
                        self.trunk_del(bridge, iface, vlan_glob)

                    # If this is a SVI for a vlan that is being deleted then delete the SVI
                    # Look for both vlan-id and vlan-raw-device to be 100% sure this is a SVI
                    elif 'vlan-id' in line:
                        vlan_id = int(line.strip().split()[1])

                        if vlan_id in vlans_to_del:
                            found_vlan_raw_device = False

                            for line2 in iface.content:
                                if 'vlan-raw-device' in line2:
                                    found_vlan_raw_device = True
                                    break

                            if found_vlan_raw_device:
                                iface.del_interface()
                                break

    def access_add(self, bridge, iface, vlan):

        # add this interface to bridge bridge-ports
        bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)

        # add this vlan to bridge bridge-vids
        bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan, give_feedback=False)

        # add bridge-access for this interface
        iface.conf_key_value_solo(True, 'bridge-access', vlan)

    def access_del(self, bridge, iface, vlan):

        # Remove this interface from bridge bridge-ports
        # We must do this else the interface would instantly become a trunk
        if ifname_in_etc_network_interfaces('bridge'):
            bridge = EtcNetworkInterface('bridge', self.ifquery_all)
            bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name)

        # remove the bridge-access line for this interface
        iface.conf_key_value_solo(False, 'bridge-access', vlan)

    def trunk_add(self, bridge, iface, vlan_glob):

        # add this interface to bridge bridge-ports
        bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)

        if vlan_glob:
            # add these vlans to the trunk bridge-vids
            iface.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob)

            # add these vlans to bridge bridge-vids
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob, give_feedback=False)

    def trunk_del(self, bridge, iface, vlan_glob):
        deconfigure_trunk = False

        # We are removing specific vlans from this trunk
        if vlan_glob:
            initial_bridge_vids = iface.get_keyword_value('bridge-vids')
            iface.conf_key_value_multiple_values(False, 'bridge-vids', vlan_glob)
            final_bridge_vids = iface.get_keyword_value('bridge-vids')

            # If we have removed the last bridge-vid then this
            # interface should no longer be a trunk
            if initial_bridge_vids and not final_bridge_vids:
                deconfigure_trunk = True

        # If the user did a del without specifying any vlans
        # then this interface should no longer be a trunk
        else:
            deconfigure_trunk = True

        if deconfigure_trunk:
            # Remove this interface from bridge bridge-ports
            if ifname_in_etc_network_interfaces('bridge'):
                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name)

            iface.conf_key_value_multiple_values(False, 'bridge-vids', '', give_feedback=False)

    def svi_add(self, bridge, iface, vlan):
        re_bridge_sub_int = re.match('^bridge\.(\d+)$', iface.name)
        re_vlan_number = re.match('^vlan\-?(\d+)$', iface.name.lower())

        if re_bridge_sub_int:
            sub_int = int(re_bridge_sub_int.group(1))
            if vlan is not None and sub_int != vlan:
                print "ERROR: The bridge sub-interface %s must be in vlan %d, it cannot be in vlan %d" % (iface.name, sub_int, vlan)
            else:
                bridge.conf_key_value_multiple_values(True, 'bridge-vids', sub_int, give_feedback=False)

        elif re_vlan_number:
            vlan = int(re_vlan_number.group(1))
            iface.conf_key_value_solo(True, 'vlan-raw-device', 'bridge')
            iface.conf_key_value_solo(True, 'vlan-id', vlan)
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan, give_feedback=False)

        elif vlan:
            iface.conf_key_value_solo(True, 'vlan-raw-device', 'bridge')
            iface.conf_key_value_solo(True, 'vlan-id', vlan)
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan, give_feedback=False)
        else:
            print "ERROR: A vlan-interface may be named 'Vlan10, Vlan20, etc' or you can specify\n"\
                  "the vlan number via 'net add vlan-interface FOO vlan 10'\n\n"

    def svi_del(self, bridge, iface, vlan):
        iface.del_interface()

    def vni_add(self, bridge, iface, vlan):
        re_vni_number = re.match('^vni\-?(\d+)$', iface.name.lower())

        if re_vni_number:
            vlan = int(re_vni_number.group(1))
            iface.conf_key_value_solo(True, 'vxlan-id', vlan)
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan, give_feedback=False)
            bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)

        elif vlan:
            iface.conf_key_value_solo(True, 'vxlan-id', vlan)
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan, give_feedback=False)
            bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)
        else:
            print "ERROR: A vni may be named 'vni10, vni-20, etc' or you can specify\n"\
                  "the vlan number via 'net add vni FOO vlan 10'\n\n"

    def vni_del(self, bridge, iface, vlan):
        iface.del_interface()

        if ifname_in_etc_network_interfaces('bridge'):
            bridge = EtcNetworkInterface('bridge', self.ifquery_all)
            bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name, give_feedback=False)

    def keywords_configured(self, args):
        """
        Return True if a keyword under an interface is configured

        We use this to tell if the user did:
            net add int swp1
            net del int swp2
        """
        for keyword in all_keywords:
            if args.get(keyword):
                return True
        return False

    def get_iface(self, ifname):
        if not self.ifquery_all:
            self.ifquery_all = check_output(['/sbin/ifquery', '-a'], stderr=STDOUT)
            self.etc_network_interfaces_init_mod_time = os.path.getmtime('/etc/network/interfaces')

        if ifname in self.ifaces:
            iface = self.ifaces[ifname]
            iface.load_iface_state()
        else:
            self.ifaces[ifname] = EtcNetworkInterface(ifname, self.ifquery_all)
            iface = self.ifaces[ifname]

        return iface

    def eval_command_line_args(self, cli):
        """
        - args is a dictionary produced by NetworkDocopt based on command line input from the user
        - Create an EtcNetworkInterface object for the interface we wish to modify
        - Find the Command object for the command keyword specified by the user (address, bridge-ports, etc)
        - Call the appropriate conf_XXX() function for this Command
        """
        args = cli.args
        ifnames = []

        # If we are being asked to modify an interface we need to verify
        # that /etc/network/interfaces has not changed. ifquery_all will be None
        # initialy and post "net commit", it will be non None once the user
        # has started staging changes.
        if args.get('add') or args.get('del'):
            if self.ifquery_all:
                if self.etc_network_interfaces_changed():
                    return
            else:
                self.ifquery_all = check_output(['/sbin/ifquery', '-a'], stderr=STDOUT)
                self.etc_network_interfaces_init_mod_time = os.path.getmtime('/etc/network/interfaces')

            ifname = args.get('<interface>')

            # Most (add|del) commands also have <interface> but not all of
            # them ('add vlans 10-20' for example
            if ifname:

                # If <interface> is a glob then populate ifnames with all of the
                # interface names produced by that glob else populate ifnames with <interface>
                if ifname_is_glob(str(ifname)):
                    ifnames = ifname_expand_glob(str(ifname))
                else:
                    if not ifname_is_valid(ifname) or ifname_is_globish(ifname):
                        return
                    ifnames = [ifname, ]

                if 'bridge' in ifnames:
                    print "ERROR: 'bridge' is a restricted interface name for net.\nUse 'net add bridge ...' to configure the bridge interface"
                    ifnames.remove('bridge')

            else:
                # "net add bridge" is an exception, we do not allow the user
                # to specify the bridge name so populate ifnames with 'bridge'
                if args.get('bridge'):
                    ifnames = ['bridge', ]

        if args.get('add') or args.get('del'):
            bridge = None
            vlan_glob = args.get('<number-range-list>')

            # 'add' for all of these require us to create a vlan-aware bridge
            # 'del' requires an EtcNetworkInterface object for the bridge (if it exists)
            if (args.get('bridge') or
                args.get('bridge-access') or
                args.get('bridge-trunk') or
                args.get('bridge-vids') or
                args.get('vlan') or
                args.get('vlan-interface') or
                args.get('vni')):

                if args.get('add'):

                    if ifname_in_etc_network_interfaces('bridge'):
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                    else:
                        # create an interface called bridge
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                        bridge.add_interface()

                    # make the bridge interface vlan aware
                    bridge.conf_key_value_solo(True, 'bridge-vlan-aware', 'yes', give_feedback=False)

                elif args.get('del'):
                    if ifname_in_etc_network_interfaces('bridge'):
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)

            # iface (add|del) vlan <number-range-list>
            if args.get('vlan') and vlan_glob and not args.get('bridge-trunk'):
                if args.get('add'):
                    self.vlan_add(bridge, vlan_glob)
                elif args.get('del'):
                    self.vlan_del(bridge, vlan_glob)

            for ifname in ifnames:
                iface = EtcNetworkInterface(ifname, self.ifquery_all)

                # If we are doing an "add" to an interface that does
                # not exist, add it automatically
                if args.get('add') and not iface.exists():
                    iface.add_interface()

                    if args.get('vlan-interface'):
                        vlan = args.get('<0-4096>')
                        self.svi_add(bridge, iface, vlan)

                    elif args.get('vni'):
                        vlan = args.get('<0-4096>')
                        self.vni_add(bridge, iface, vlan)

                # The bridge-trunk, vlan-interface, vni-interfce, etc keywords are special, these
                # are not from an ifupdown2 addon.  They are keywords that we
                # explicitly added in the __doc__ so must handle these here and not
                # in the for loop for the addons.
                #
                # bridge-access is from an ifupdown2 addon but we want to overwrite it and
                # treat it as a macro so we handle it here instead of via the addon for loop.

                # iface add interface <interface> access vlan <0-4096>
                if args.get('bridge-access'):
                    vlan = args.get('<0-4096>')

                    if args.get('add'):
                        self.access_add(bridge, iface, vlan)
                    elif args.get('del'):
                        self.access_del(bridge, iface, vlan)

                #
                # iface (add|del) interface <interface> bond-master <interface-bond>
                #
                #elif args.get('bond-master'):
                #    bond = EtcNetworkInterface(args.get('<interface-bond>'), self.ifquery_all)
                #
                #    if args.get('add'):
                #        if not bond.exists():
                #            bond.add_interface()
                #        bond.conf_key_value_multiple_values(True, 'bond-slaves', ifname)
                #
                #    elif args.get('del'):
                #        if bond.exists():
                #            bond.conf_key_value_multiple_values(False, 'bond-slaves', ifname)
                #        else:
                #            print "ERROR: bond %s does not exist" % bond

                # The user can configure a trunk via bridge-trunk or bridge-vids.
                # bridge-vids can be configured on the bridge interface itself though
                # which is used to add support for a vlan, not configure a trunk.  Check
                # to see if bridge-vids is being configured but not on the 'bridge' interface
                #
                # iface (add|del) interface <interface> bridge-trunk [vlans <number-range-list>]
                elif args.get('bridge-trunk') or (args.get('bridge-vids') and not args.get('bridge')):
                    if args.get('add'):
                        self.trunk_add(bridge, iface, vlan_glob)
                    elif args.get('del'):
                        self.trunk_del(bridge, iface, vlan_glob)

                # iface (add|del) vlan-interface <interface> vlan <0-4096>
                elif args.get('vlan-interface') and not self.keywords_configured(args):
                    vlan = args.get('<0-4096>')

                    if args.get('add'):
                        pass
                    elif args.get('del'):
                        self.svi_del(bridge, iface, vlan)

                # iface (add|del) vni <interface> vlan <0-4096>
                elif args.get('vni') and not self.keywords_configured(args):
                    vlan = args.get('<0-4096>')

                    if args.get('add'):
                        pass
                    elif args.get('del'):
                        self.vni_del(bridge, iface, vlan)


                # word-around for the word 'vrf'.  vrf is an interface type
                #   iface (add|del) vrf BLUE
                #
                # And it can be used to put an interface in a vrf
                #   iface (add)del int swp1 vrf BLUE
                #
                # If the length is 4 we know it is 'iface (add|del) vrf BLUE' so handle
                # that here as a one of
                elif len(cli.argv_expanded) == 4 and cli.argv_expanded[2] == 'vrf':

                    # We added it earlier so ignore 'add' here
                    if cli.argv_expanded[1] == 'add':
                        pass
                    elif cli.argv_expanded[1] == 'del':
                        iface.del_interface()

                # A "normal" (non macro) command that comes straight from importing an ifupdown2 addon
                else:

                    # Find the Command object that matches
                    for cmd in addons_cmds:
                        if not cmd.deprecated and args.get(cmd.keyword):
                            ifaces_to_add = []

                            # Another work-around for the 'vrf' keyword collision
                            if cmd.keyword == 'vrf' and len(cli.argv_expanded) >= 3 and cli.argv_expanded[2] == 'vrf':
                                continue

                            # If we are doing an "add" to an interface that does
                            # not exist, add it automatically
                            if args.get('add') and not iface.exists():
                                iface.add_interface()

                            if isinstance(cmd, CommandAddDelMultiLine):
                                iface.conf_key_value_multiple_lines(args.get('add'), cmd.keyword, cmd.value_from_args(args))
                            elif isinstance(cmd, CommandAddDelMultiValue):
                                ifaces_to_add = iface.conf_key_value_multiple_values(args.get('add'), cmd.keyword, cmd.value_from_args(args))
                            elif isinstance(cmd, CommandSetClear):
                                iface.conf_key_value_solo(args.get('add'), cmd.keyword, cmd.value_from_args(args))
                            else:
                                raise Exception("command %s is an invalid type %s" % (cmd, type(cmd)))

                            # Automagically add the interfaces listed under bridge-ports, bond-slaves, etc
                            self.auto_add_interfaces(ifaces_to_add)
                            break
                    else:
                        # If we are here the user did one of the following
                        #  net add int swp20
                        #  net del int swp20
                        #
                        # We already handled the "add" scenario prior to the for loop above
                        # "del" we handle here
                        if args.get('add'):
                            if not iface.exists():
                                iface.add_interface()
                        elif args.get('del'):
                            iface.del_interface()

    def etc_network_interfaces_changed(self):
        """
        Return True if some other user modified /etc/network/interfaces
        """

        # If this is the case we haven't started making changes
        # to /etc/network/interfaces yet so return True
        if self.etc_network_interfaces_init_mod_time is None:
            return False

        etc_network_interfaces_mod_time = os.path.getmtime('/etc/network/interfaces')

        if self.etc_network_interfaces_init_mod_time != etc_network_interfaces_mod_time:
            print "/etc/network/interfaces was modified by another user."
            print_pending_add_del_commands()
            return True
        return False

    def _save_or_show_pending(self, persist, verbose):
        """
        if persist is True
        - write changes to /etc/network/interfaces
        - delete various tmp files such as .del and .original
        - 'ifreload -a' to apply changes

        else
        - write changes to /var/run/nclu/iface/interfaces.tmp
        """

        if self.etc_network_interfaces_changed():
            return False

        if persist:
            target_filename = '/etc/network/interfaces'
        else:
            target_filename = '/var/run/nclu/iface/interfaces.tmp'
            copyfile('/etc/network/interfaces', target_filename)

        iface_files = {}
        iface_delete = {}

        # Make a first pass to:
        # - find all interfaces with a .del file
        # - remove the .original files
        for ifname in os.listdir('/var/run/nclu/iface/'):
            filename = "/var/run/nclu/iface/%s" % ifname

            if filename.endswith('.del'):
                ifname = ifname[:-4]
                iface_delete[ifname] = True

                if persist:
                    print "%s will be removed from /etc/network/interfaces" % ifname
                    os.remove(filename)

                    if os.path.isfile("/var/run/nclu/iface/%s" % ifname):
                        os.remove("/var/run/nclu/iface/%s" % ifname)

            elif filename.endswith('.original'):
                if persist:
                    os.remove(filename)

        # Make a second pass to load the content for interfaces that have been modified
        # but are not being deleted
        for ifname in os.listdir('/var/run/nclu/iface/'):

            # ignore these
            if (ifname in ('interfaces_for_diff', 'interfaces.tmp') or
                ifname.endswith('.del') or ifname.endswith('.original')):
                continue

            filename = "/var/run/nclu/iface/%s" % ifname

            # This should never happen but ignore it if the user creates
            # a directory, etc in /var/run/nclu/iface/
            if not os.path.isfile(filename):
                print "ERROR: commit_pending() found %s which is not a file (ignoring)" % filename
                continue

            with open(filename, 'r') as fh:
                iface_files[ifname] = fh.readlines()

        # Sanity check that mako templates are not in use
        with open('/etc/network/interfaces', 'r') as fh:
            for line in fh.readlines():
                line = line.strip()
                if line.startswith('<%') or line.startswith('% '):
                    print "/etc/network/interfaces contains mako templates, net does not support mako templates"
                    return False

        # r+ opens the file for reading and writing
        with open(target_filename, 'r+') as fh:

            # Read in all lines and the reset the fh pointer back to
            # the start of the file
            lines = fh.readlines()
            fh.seek(0)

            state = 'WRITE'
            prev_line = None
            lines_to_write = []

            # Now write all of the lines back to the file but update
            # the stanzas for interfaces with /var/run/nclu/iface/ files
            #
            # This will maintain the original order of the stanzas
            for line in lines:
                line = standardize_indent(line).rstrip()
                re_auto = re.search("^\s*auto\s+(\S+)\s*$", line)

                if re_auto:
                    ifname = re_auto.group(1)

                    # Look to see if any of the interfaces in iface_files need to be
                    # written before this interface in order to keep them sorted
                    # in /etc/network/interfaces
                    ifaces_to_add_here = []
                    sorted_interfaces = sort_ifaces(iface_files.keys() + [ifname, ])

                    for x in sorted_interfaces:
                        if x == ifname:
                            break
                        ifaces_to_add_here.append(x)

                    # print "%s: sorted_interface %s" % (ifname, sorted_interfaces)
                    # print "%s: interfaces to add here %s" % (ifname, ifaces_to_add_here)

                    if ifaces_to_add_here:
                        # We are at an 'auto' line and we need to insert interfaces
                        # prior to that auto line.  That part is easy but what if
                        # there are comments before this auto line?  We need to
                        # insert our interfaces before those comments.
                        comments = []

                        for x in reversed(lines_to_write):
                            if x.startswith('#'):
                                comments.append(x)
                            else:
                                break

                        # Remove the comments for this 'auto'
                        if comments:
                            comments_to_chop = -1 * len(comments)
                            lines_to_write = lines_to_write[:comments_to_chop]

                        # Add the interfaces that need to go before this auto
                        for x in ifaces_to_add_here:
                            write_new_interface(lines_to_write, prev_line, x, iface_files[x], persist)
                            del iface_files[x]

                        # Restore the comments for this 'auto'
                        for x in reversed(comments):
                            lines_to_write.append(x)

                    # Do not write any of the config lines for an interface we are trying to delete
                    if ifname in iface_delete:
                        log.debug("%s: remove from %s" % (ifname, target_filename))
                        del iface_delete[ifname]

                        if state == 'WRITE':
                            # log.debug("WRITE->SKIP: %s" % line)
                            state = 'SKIP'

                    # This is an interface that has been modified via iface. Write the config lines
                    # from the iface file to target_filename
                    elif ifname in iface_files:
                        log.debug("%s: update in %s" % (ifname, target_filename))

                        for iface_line in iface_files[ifname]:
                            iface_line = iface_line.rstrip()
                            log.debug("%s: write '%s'" % (ifname, iface_line))
                            lines_to_write.append(iface_line + '\n')
                        lines_to_write.append('\n')

                        if persist:
                            os.remove("/var/run/nclu/iface/%s" % ifname)
                        del iface_files[ifname]

                        if state == 'WRITE':
                            # log.debug("WRITE->SKIP: %s" % line)
                            state = 'SKIP'

                    # This is an interface that has not been modified via iface
                    else:

                        if state == 'SKIP':
                            # log.debug("SKIP->WRITE: %s" % line)
                            state = 'WRITE'

                elif line.startswith('source '):
                    if state == 'SKIP':
                        # log.debug("SKIP->WRITE: %s" % line)
                        state = 'WRITE'

                # If we are parsing comments then set state back to WRITE
                elif state == 'SKIP' and line.startswith('#'):
                    state = 'WRITE'
                    # log.debug("SKIP->WRITE: %s" % line)

                if state == 'WRITE':
                    lines_to_write.append(line + '\n')
                elif state == 'SKIP':
                    pass
                else:
                    raise Exception("state %s is invalid" % state)

                prev_line = line

            # Write any remaining iface files to target_filename
            # These will be sorted in human readable order
            # print "%s: ifaces left to add %s\n" % (ifname, iface_files.keys())
            for ifname in sort_ifaces(iface_files.keys()):
                write_new_interface(lines_to_write, prev_line, ifname, iface_files[ifname], persist)

            # Now that we have constructed what we want to write, write it.
            prev_line = None
            number_lines_to_write = len(lines_to_write)

            for (i, line) in enumerate(lines_to_write):
                if line.strip():
                    fh.write(line)
                else:
                    # Do not print multiple blank lines in a row
                    if prev_line is None or prev_line.strip():

                        # Do not print a blank line at the end
                        if i < number_lines_to_write - 1:
                            fh.write(line)
                prev_line = line

            # If we deleted lines we need to truncate the rest of the fh so that we
            # do not end up with the lines at the end printed twice
            fh.truncate()

        if persist:
            if verbose:
                print "Updating /etc/network/interfaces and applying changes via 'ifreload -va'"
                output = check_output(['/sbin/ifreload', '-va'], stderr=STDOUT)

            else:
                if self.show_linux_command:
                    print "Updating /etc/network/interfaces and applying changes via 'ifreload -a'"
                output = check_output(['/sbin/ifreload', '-a'], stderr=STDOUT)

            if output:
                print output

        return True
