#!/usr/bin/env python

"""
Usage:

    # Linux Basics
    netmisc add hostname <text>
    netmisc show hostname
    netmisc show version

Options:
    help   : show this screen and exit
"""

from network_docopt import NetworkDocopt, get_network_docopt_info
from subprocess import check_output, STDOUT
import json
import logging
import logging.handlers
import os
import re
import shlex
import sys

try:
    from netshowlib import interface_related, check as vendor_check, \
        ShowInterfaces, CumulusShowInterfaces, \
        ShowNeighbors, CumulusShowNeighbors, \
        ShowSystem, CumulusShowSystem
    netshow_present = True
except ImportError:
    netshow_present = False

log = logging.getLogger('netd')


def valid_hostname(hostname):
    if len(hostname) < 1 or len(hostname) > 63:
        return False

    for x in hostname:
        if x.isalpha() or x.isdigit() or x == '-':
            pass
        else:
            return False

    return True


def edit_hostname(new_hostname):
    filename = '/var/run/nclu/netmisc/execlist'
    lines = []

    if not valid_hostname(new_hostname):
        print "ERROR: %s is an invalid hostname, hostnames may only contain letters, numbers, or -" % new_hostname
        return

    if os.path.exists(filename):
        with open(filename, 'r') as fh:
            for line in fh.readlines():

                # If the user did multiple "net add hostname FOO" back to back we
                # do not want to add those commands multiple times so delete all
                # of the hostname related commands that are waiting in the execlist
                if (line.startswith('hostname') or
                    '" /etc/hostname' in line or
                    '" /etc/hosts' in line or
                    'You must re-login for your bash prompt to reflect the hostname change' in line):
                    pass
                else:
                    lines.append(line)

    with open(filename, 'w') as fh:
        for line in lines:
            line = line.rstrip()
            fh.write(line + '\n')

        old_hostname = check_output(['/bin/hostname']).strip()
        fh.write('hostname %s\n' % new_hostname)

        # Use sed to overwrite /etc/hostname so that we do not use > which requires shell=True
        fh.write('sed -i "s/.*/%s/" /etc/hostname\n' % new_hostname)
        fh.write('sed -i "/127\.0\.1\.1/ s/%s/%s/" /etc/hosts\n' % (old_hostname, new_hostname))
        fh.write('hostnamectl set-hostname %s\n' % new_hostname)
        fh.write('echo "You must re-login for your bash prompt to reflect the hostname change"\n')


def build_doc_string():
    global __doc__

    # If netshow is present wrapper it
    if netshow_present:
        __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # interface show commands
    netmisc show interface [all|bonds|bondmems|<interface>] [mac|json]
    netmisc show l2 [access|bridges|trunks] [mac|json]
    netmisc show lldp [json]
    netmisc show system""")

    # If cl-netstat is present wrapper it
    if os.path.exists('/usr/cumulus/bin/cl-netstat'):
        __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # interface counters
    netmisc clear counters
    netmisc show counters [json]""")

        __doc__ = __doc__.replace('show this screen and exit', """show this screen and exit
    access     : summary of physical ports with l2 or l3 config
    bondmems   : summary of bond members
    bonds      : summary of bonds
    bridges    : summary of ports with bridge members
    counters   : summary of physical port counters.
    interface  : summary info of all interfaces
    l2         : summary of access, trunk and bridge interfaces
    l3         : summary of ports with an IP.
    lldp       : physical device neighbor information
    mgmt       : summary of mgmt ports
    system     : system information
    trunks     : summary of trunk interfaces
    all        : everything
    json       : print output in json
    mac        : show inteface MAC in output
            """)


class ConfigWrapper(object):

    def __init__(self, show_linux_command):
        self.show_linux_command = show_linux_command

    def abort_pending(self):
        if os.path.exists('/var/run/nclu/netmisc/'):
            for filename in os.listdir('/var/run/nclu/netmisc/'):
                full_filename = os.path.join('/var/run/nclu/netmisc/', filename)

                if os.path.isdir(full_filename):
                    rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def commit_pending(self, verbose):
        full_filename = '/var/run/nclu/netmisc/execlist'
        files_modified = []

        if os.path.exists(full_filename):
            with open(full_filename, 'r') as fh:
                for line in fh.readlines():
                    cmd = shlex.split(line.strip())

                    if verbose:
                        if self.show_linux_command:
                            print "exec: %s" % ' '.join(cmd)
                        output = check_output(cmd, stderr=STDOUT)
                        print output
                    else:
                        check_output(cmd, stderr=STDOUT)
            os.remove(full_filename)

        # This needs some work
        return files_modified

    def show_pending(self, print_errors=True):
        full_filename = '/var/run/nclu/netmisc/execlist'
        diff_pending = False

        if os.path.exists(full_filename):
            with open(full_filename, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('hostname ') and print_errors:
                        print "\nWe will run the following commands to change the hostname"
                        print "========================================================="

                    print line
                    diff_pending = True
            print '\n'

        return diff_pending

    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user
        """
        args = cli.args

        if args.get('show'):

            # We do get here because 'counters' will not be in args if we are on a
            # box without cl-netstat
            if args.get('counters'):
                cmd = ['/usr/cumulus/bin/cl-netstat', ]

                if args.get('json'):
                    cmd.append('--json')

                print check_output(cmd)

            elif args.get('hostname'):
                sys.stdout.write(check_output(('/bin/hostname',)))

            elif args.get('version'):
                with open('/etc/lsb-release', 'r') as fh:
                    for line in fh.readlines():
                        sys.stdout.write(line)

            # netshow
            else:
                vendor = vendor_check()
                show = None

                # 'access', 'bridges', 'bonds', 'bondmems', 'mgmt', 'l2', 'l3', 'phy', 'trunks', 'interface'
                if interface_related(cli):
                    if vendor == 'cumulus':
                        show = CumulusShowInterfaces(cli)
                    else:
                        show = ShowInterfaces(cli)

                elif cli.get('system'):
                    if vendor == 'cumulus':
                        show = CumulusShowSystem(cli)
                    else:
                        show = ShowSystem(cli)

                elif cli.get('lldp'):
                    if vendor == 'cumulus':
                        show = CumulusShowNeighbors(cli)
                    else:
                        show = ShowNeighbors(cli)

                if show:
                    print show.run()
                else:
                    print "ERROR: could not determine what data to show"

        elif args.get('clear'):
            # We do get here because 'counters' will not be in args if we are on a
            # box without cl-netstat
            if args.get('counters'):
                cmd = ['/usr/cumulus/bin/cl-netstat', '--clear']
                if self.show_linux_command:
                    print "exec '%s'\n" % ' '.join(cmd)
                print check_output(cmd)

        # (add|del)
        else:
            if args.get('hostname'):
                edit_hostname(args.get('<text>'))
