#!/usr/bin/python

"""
Usage:
    # L3 clear commands
    quagga clear bgp (<bgppeer>|all) [soft in|soft out]
    quagga clear bgp prefix <ipv4/prefixlen>
    quagga clear ospf interface [<interface>]

    # L3 show commands
    quagga show bgp [summary]
    quagga show bgp (<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
    quagga show bgp ipv4 unicast [<ipv4>|<ipv4/prefixlen>|summary] [json]
    quagga show bgp ipv6 unicast [<ipv6>|<ipv6/prefixlen>|summary] [json]
    quagga show bgp neighbor [<bgppeer>]
    quagga show bgp update-groups
    quagga show ospf [json]
    quagga show ospf database (asbr-summary|external|network|nssa-external|opaque-area|opaque-as|opaque-link|router|summary) [<ipv4>|adv-router <ipv4-adv-router>|self-originate]
    quagga show ospf database [max-age|self-originate]
    quagga show ospf interface [<interface>] [json]
    quagga show ospf neighbor [all|<interface>|<ipv4>|detail] [json]
    quagga show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|ospf|static|summary]
    quagga show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes

    # L3 debug commands
    quagga add       debug bgp bestpath <ip/prefixlen>
    quagga del       debug bgp bestpath [<ip/prefixlen>]
    quagga add       debug bgp keepalives (<ip>|<interface>)
    quagga del       debug bgp keepalives [<ip>|<interface>]
    quagga add       debug bgp neighbor-events (<ip>|<interface>)
    quagga del       debug bgp neighbor-events [<ip>|<interface>]
    quagga add       debug bgp updates prefix <ip/prefixlen>
    quagga del       debug bgp updates prefix [<ip/prefixlen>]
    quagga add       debug bgp zebra prefix <ip/prefixlen>
    quagga del       debug bgp zebra prefix [<ip/prefixlen>]
    quagga (add|del) debug ospf ism [status|events|timers]
    quagga (add|del) debug ospf lsa [generate|flooding|install|refresh]
    quagga (add|del) debug ospf nsm [status|events|timers]
    quagga (add|del) debug ospf packet (hello|dd|ls-request|ls-update|ls-ack|all) [send|recv|detail]
    quagga (add|del) debug ospf zebra [interface|redistribute]

    # L3 global configuration
    quagga add       router-id <ipv4>
    quagga del       router-id [<ipv4>]
    quagga (add|del) ip route <ipv4/prefixlen> (<ipv4>|<interface>|Null0|blackhole|reject) [<1-255>] [vrf <text>]
    quagga (add|del) ipv6 route <ipv6/prefixlen> (<ipv6>|<interface>) [blackhole|reject] [<1-255>] [vrf <text>]
    quagga (add|del) ipv6 route <ipv6/prefixlen> <ipv6> <interface> [blackhole|reject] [<1-255>] [vrf <text>]

    # L3 BGP 'global' configuration
    quagga (add|del) router bgp <1-4294967295>
    quagga (add|del) bgp autonomous-system <1-4294967295>
    quagga (add|del) bgp always-compare-med
    quagga (add|del) bgp bestpath as-path multipath-relax [as-set|no-as-set]
    quagga (add|del) bgp neighbor <bgppeer> capability extended-nexthop
    quagga add       bgp neighbor <bgppeer> description <wildcard>
    quagga del       bgp neighbor <bgppeer> description [<wildcard>]
    quagga (add|del) bgp neighbor <bgppeer> ebgp-multihop [<1-255>]
    quagga add       bgp neighbor <bgppeer> password <text>
    quagga del       bgp neighbor <bgppeer> password [<text>]
    quagga (add|del) bgp neighbor (<ip>|<interface>|<bgppeer>) peer-group <peergroup>
    quagga (add|del) bgp neighbor (<ip>|<interface>|<bgppeer>) remote-as (<1-4294967295>|external|internal)
    quagga (add|del) bgp neighbor <bgppeer> shutdown
    quagga add       bgp neighbor <bgppeer> timers <0-65535> <3-65535>
    quagga del       bgp neighbor <bgppeer> timers [<0-65535>] [<3-65535>]
    quagga add       bgp neighbor <bgppeer> timers connect <1-65535>
    quagga del       bgp neighbor <bgppeer> timers connect [<1-65535>]
    quagga add       bgp neighbor <bgppeer> update-source (<ipv4>|<ipv6>|<interface-source>)
    quagga del       bgp neighbor <bgppeer> update-source [<ipv4>|<ipv6>|<interface-source>]
    quagga (add|del) bgp neighbor <interface> interface
    quagga (add|del) bgp neighbor <interface> interface peer-group <peergroup>
    quagga (add|del) bgp neighbor <text> peer-group
    quagga add       bgp router-id <ipv4>
    quagga del       bgp router-id [<ipv4>]

    # L3 BGP per address-family configuration
    quagga (add|del) bgp [ipv4 unicast] aggregate-address <ipv4/prefixlen> [as-set] [summary-only]
    quagga (add|del) bgp ipv6 unicast aggregate-address <ipv6/prefixlen> [summary-only]
    quagga (add|del) bgp [ipv4 unicast] network <ipv4/prefixlen>
    quagga (add|del) bgp ipv6 unicast network <ipv6/prefixlen>
    quagga (add|del) bgp [ipv4 unicast] redistribute (connected|isis|kernel|ospf|pim|rip|static|table) [route-map <route-map>]
    quagga (add|del) bgp ipv6 unicast redistribute (connected|isis|kernel|ospf6|ripng|static|table) [route-map <route-map>]
    quagga add       bgp [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] <1-64>
    quagga del       bgp [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] [<1-64>]
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> activate
    quagga add       bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> allowas-in <1-10>
    quagga del       bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> allowas-in [<1-10>]
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> default-originate
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> default-originate route-map <route-map>
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> filter-list <as-path-access-list> (in|out)
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> next-hop-self [force]
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> prefix-list <prefix-list-v4> (in|out)
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> prefix-list <prefix-list-v6> (in|out)
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> remove-private-AS [all] [replace-AS]
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> route-map <route-map> (in|out)
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> route-reflector-client
    quagga (add|del) bgp [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> send-community [both|extended|standard]

    # L3 OSPF configuration
    quagga (add|del) router ospf
    quagga (add|del) ospf network <ipv4/prefixlen> area (<ipv4>|<0-4294967295>)
    quagga del       ospf network <ipv4/prefixlen>
    quagga (add|del) ospf passive-interface <interface>
    quagga add       ospf router-id <ipv4>
    quagga del       ospf router-id [<ipv4>]
    quagga add       ospf timers throttle spf <0-600000> <1-600000> <2-600000>
    quagga del       ospf timers throttle spf [<0-600000> <1-600000> <2-600000>]
    quagga add       interface <interface> ip ospf area (<ipv4>|<0-4294967295>)
    quagga del       interface <interface> ip ospf area [<ipv4>|<0-4294967295>]
    quagga add       interface <interface> ip ospf dead-interval <1-65535>
    quagga del       interface <interface> ip ospf dead-interval [<1-65535>]
    quagga add       interface <interface> ip ospf hello-interval <1-65535>
    quagga del       interface <interface> ip ospf hello-interval [<1-65535>]
    quagga add       interface <interface> ip ospf network (broadcast|non-broadcast|point-to-multipoint|point-to-point)
    quagga del       interface <interface> ip ospf network [broadcast|non-broadcast|point-to-multipoint|point-to-point]

    # as-path lists
    quagga (add|del) ip as-path access-list <text> (deny|permit) <wildcard>
    qaugga show ip as-path-access-list [<as-path-access-list>]

    # community-list
    quagga (add|del) ip community-list standard <text> (deny|permit) (local-AS|no-advertise|no-export|internet|<community>)
    quagga (add|del) ip community-list expanded <text> (deny|permit) <regex>
    quagga (add|del) ip extcommunity-list standard <text> (deny|permit) (rt|soo) <extcommunity>
    quagga (add|del) ip extcommunity-list expanded <text> (deny|permit) <regex>

    quagga show ip community-list <community-list>
    quagga show ip extcommunity-list <extcommunity-list>

    # prefix-list
    quagga (add|del) ip prefix-list <text> description <wildcard>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) (<ipv4/prefixlen>|any)
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32> le <1-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32> ge <1-32>
    quagga (add|del) ipv6 prefix-list <text> description <wildcard>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) (<ipv6/prefixlen>|any)
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128> le <1-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128> ge <1-128>
    quagga clear ip prefix-list [<prefix-list-v4>|<prefix-list-v4> <ipv4/prefixlen>]
    quagga clear ipv6 prefix-list [<prefix-list-v6>|<prefix-list-v6> <ipv6/prefixlen>]
    quagga show ip prefix-list [<prefix-list-v4>]
    quagga show ip prefix-list <prefix-list-v4> seq <1-4294967295>
    quagga show ip prefix-list <prefix-list-v4> <ipv4/prefixlen> [longer|first-match]
    quagga show ip prefix-list summary [<prefix-list-v4>]
    quagga show ip prefix-list detail [<prefix-list-v4>]
    quagga show ipv6 prefix-list [<prefix-list-v6>]
    quagga show ipv6 prefix-list <prefix-list-v6> seq <1-4294967295>
    quagga show ipv6 prefix-list <prefix-list-v6> <ipv6/prefixlen> [longer|first-match]
    quagga show ipv6 prefix-list summary [<prefix-list-v6>]
    quagga show ipv6 prefix-list detail [<prefix-list-v6>]

    # route-map
    quagga (add|del) route-map <text> (deny|permit) <1-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> call <route-map>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> continue <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> description <wildcard>
    quagga del route-map <text> permit <1-65535> on-match (goto <0-65535>|next)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match as-path <as-path-access-list>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match community <community-list> [exact-match]
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match extcommunity <extcommunity-list>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match interface <interface>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ip (address|next-hop) prefix-len <0-32>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ip (address|next-hop|route-source) prefix-list <prefix-list-v4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ipv6 address prefix-list <prefix-list-v6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ipv6 next-hop <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match local-preference <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match metric <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match origin (egp|igp|incomplete)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match peer (<ipv4>|<ipv6>|local)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match probability <0-100>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match source-protocol (bgp|connected|isis|kernel|ospf|ospf6|rip|ripng|static|system)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match tag <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set aggregator as <1-4294967295> <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set as-path exclude <1-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set as-path prepend (<1-4294967295>|last-as <1-10>)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set atomic-aggregate
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set comm-list <community-list> delete
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set community <community> [additive]
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set community none
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set extcommunity (rt|soo) <extcommunity>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set forwarding-address <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ip next-hop (<ipv4>|peer-address|unchanged)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ipv6 next-hop (global|local) <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ipv6 next-hop peer-address
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set local-preference <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set metric <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set metric-type (type-1|type-2)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set origin (egp|igp|incomplete)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set originator-id <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set src (<ipv4>|<ipv6>)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set tag <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set vpnv4 next-hop <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set weight <0-4294967295>
    quagga show route-map [<route-map>]


Options:
    activate                   : Enable the Address Family for this Neighbor
    addpath-tx-all-paths       : Use addpath to advertise all paths to a neighbor
    addpath-tx-bestpath-per-AS : Use addpath to advertise the bestpath per each neighboring AS
    advertisement-interval     : Minimum interval between sending BGP routing updates
    aggregate-address          : Configure BGP aggregate entries
    allowas-in                 : Accept as-path with my AS present in it
    always-compare-med         : Allow comparing MED from different neighbors
    area                       : An OSPF area
    asbr-summary               : ASBR summary link states
    as-override                : Override ASNs in outbound updates if aspath equals remote-as
    as-path                    : AS-path attribute
    as-set                     : Generate an AS_SET
    attribute-unchanged        : BGP attribute is propagated unchanged to this neighbor
    autonomous-system          : AS number
    bestpath                   : BGP bestpath
    bfd                        : Enables BFD support
    bgp                        : Border Gateway Protocol
    blackhole                  : Silently discard pkts when matched
    border-routers             : OSPF ABR's and ASBR's
    broadcast                  : An OSPF broadcast multi-access network
    capability                 : Advertise capability to the peer
    client-to-client           : Configure client to client route reflection
    cluster-id                 : Configure Route-Reflector Cluster-id
    community-list             : Add a community list entry
    compare-routerid           : Compare router-id for identical EBGP paths
    confed                     : Compare path lengths including confederation sets & sequences in selecting a route
    confederation              : AS confederation parameters
    connected                  : Connected routes (directly attached subnet or host)
    dampening                  : Enable route-flap dampening
    database                   : OSPF database
    dead-interval              : Interval after which a neighbor is declared dead
    debug                      : Enable debugs
    default-originate          : Originate default route to this neighbor
    description                : Neighbor specific description
    deterministic-med          : Pick the best-MED path among paths advertised from the neighboring AS
    disable-connected-check    : one-hop away EBGP peer using loopback address
    disable-ebgp-connected-route-check : Disable checking if nexthop is connected on ebgp sessions
    distribute-list            : Filter updates to/from this neighbor
    dont-capability-negotiate  : Do not perform capability negotiation
    ebgp-multihop              : Allow EBGP neighbors not on directly connected networks
    enforce-first-as           : Enforce the first AS for EBGP routes
    enforce-multihop           : Enforce EBGP neighbors perform multihop
    event                      : event information
    extcommunity-list          : Add a extended community list entry
    external                   : An external BGP peer, external OSPF LSA, etc
    fast-external-failover     : Immediately reset session if a link to a directly connected external peer goes down
    filter-list                : Apply an as-path access-list
    flooding                   : LSA Flooding
    generate                   : LSA Generation
    graceful-restart           : Graceful restart capability parameters
    hello-interval             : Time between HELLO packets
    ibgp                       : An internal BGP peer
    ignore                     : Ignore as-path length in selecting a route
    import-table               : import routes from non-main kernel table
    in                         : apply inbound
    install                    : LSA Install/Delete
    interface                  : Interface
    internal                   : An internal BGP peer
    ip                         : Internet Protocol version 4
    ipv4                       : Internet Protocol version 4
    ipv6                       : Internet Protocol version 6
    isis                       : Intermediate System to Intermediate System (IS-IS)
    ism                        : Interface State Machine
    kernel                     : Kernel routes (not installed via the zebra RIB)
    local-as                   : Specify a local-as number
    log-neighbor-changes       : Log neighbor up/down and reset reason
    longer-prefixes            : Show route matching the specified Network/Mask pair only
    lsa                        : Link State Advertisement
    max-age                    : LSAs in MaxAge list
    maximum-paths              : Forward packets over multiple paths
    maximum-prefix             : Maximum number of prefix accept from this peer
    max-med                    : Advertise routes with max-med
    med                        : MED attribute
    mroute                     : Configure static unicast route into MRIB for multicast RPF lookup
    multicast                  : Multicast options
    multicast-routing          : Enable IP multicast forwarding
    multipath                  : BGP paths flagged as multipath
    multipath-relax            : Allow load sharing across routes that have different AS paths (but same length)
    neighbor                   : An OSPF, BGP, etc neighbor
    network                    : An ipv4 or ipv6 network
    next-hop-self              : Disable the next hop calculation for this neighbor
    nht                        : Next Hop Tracking
    no-as-set                  : Do not generate an AS_SET
    non-broadcast              : An OSPF NBMA network
    nsm                        : Neighbor State Machine
    nssa-external              : NSSA external link state
    nssa                       : OSPF nssa information
    opaque-area                : Link area Opaque-LSA
    opaque-as                  : Link AS Opaque-LSA
    opaque-link                : Link local Opaque-LSA
    ospf                       : Open Shortest Path First (OSPFv2)
    out                        : apply outbound
    override-capability        : Override capability negotiation result
    packet                     : packet
    passive                    : Don't send open messages to this neighbor
    password                   : Set a password
    peer-group                 : Member of the peer-group
    pim                        : Protocol Independent Multicast
    point-to-multipoint        : An OSPF point-to-multipoint network
    point-to-point             : An OSPF point-to-point network
    port                       : Neighbor's TCP port
    prefix-list                : Filter updates to/from this neighbor
    protocol                   : Filter routing info exchanged between zebra and protocol
    redistribute               : Redistribute information from another routing protocol
    refresh                    : LSA Refresh
    reject                     : Emit an ICMP unreachable when matched
    remote-as                  : Specify a BGP neighbor
    remove-private-AS          : Remove private ASNs in outbound updates
    replace-AS                 : Replace private ASNs with our ASN in outbound updates
    rip                        : Routing Information Protocol (RIP)
    route-map                  : route-map
    route-reflector            : Allow modifications made by out route-map
    route-reflector-client     : Configure a neighbor as Route Reflector client
    router-id                  : router identifier
    router                     : L3 routing protocols or OSPF LSA type
    route-server-client        : Configure a neighbor as Route Server client
    route                      : static routes
    self-originate             : Self-originated link states
    send-community             : Send Community attribute to this neighbor
    shutdown                   : Administratively shut down this neighbor
    soft-reconfiguration       : Per neighbor soft reconfiguration
    soft                       : Soft reconfig inbound and outbound updates
    solo                       : Solo peer - part of its own update group
    spf                        : OSPF SPF timers
    static                     : Statically configured routes
    strict-capability-match    : Strict capability negotiation match
    summary                    : Network summary link states
    summary-only               : Filter more specific routes from updates
    table-map                  : BGP table to RIB route download filter
    table                      : Non-main Kernel Routing Table
    tag                        : Set tag for this route
    throttle                   : Throttling adaptive timer
    timers                     : Adjust routing timers
    timers                     : BGP per neighbor timers
    ttl-security               : Specify the maximum number of hops to the BGP peer
    unsuppress-map             : Route-map to selectively unsuppress suppressed routes
    update-source              : Source of routing updates
    vpnv4                      : Address family
    vrf                        : Virtual routing and forwarding
    weight                     : Set default weight for routes from this neighbor
    zebra                      : Zebra information

  assert                 : PIM interface assert
  assert-internal        : PIM interface internal assert state
  assert-metric          : PIM interface assert metric
  assert-winner-metric   : PIM interface assert winner metric
  designated-router      : PIM interface designated router
  hello                  : PIM interface hello information
  join                   : PIM interface join information
  jp-override-interval   : PIM interface J/P override interval
  lan-prune-delay        : PIM neighbors LAN prune delay parameters
  local-membership       : PIM interface local-membership
  rpf                    : PIM cached source rpf information
  secondary              : PIM neighbor addresses
  upstream               : PIM upstream information
  upstream-join-desired  : PIM upstream join-desired
  upstream-rpf           : PIM upstream source rpf
  groups      : IGMP groups information
  parameters  : IGMP parameters information
  querier     : IGMP querier information
  sources     : IGMP sources information
  count       :  Route and packet count data
  deny           : deny in routing policy
  permit         : permit in routing policy
  <regex>        : A regular expression
  internet       : BGP community, matches all routes
  local-AS       : BGP community, restrict routes to our confederations sub-AS
  no-advertise   : BGP community, do not advertise to anyone
  no-export      : BGP community, do not advertise to eBGP peer
  <community>    : XX:YY where XX and YY are numbers in the range 0-65535
  <extcommunity> : ASN:nn_or_IP-address:nn VPN extended community
"""

# Syntax for access-list...we decided to leave these out since prefix-lists
# are a much better way for filtering routes.
'''
    # access-list
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> host <ipv4-dest>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any host <ipv4-dest>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> host <ipv4-dest>
    quagga (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) <ipv4-source> [<ipv4-source-wildcard>]
    quagga (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) (any|host <ipv4-source>)
    quagga (add|del) access-list <text> (deny|permit) (any|<ipv4/prefixlen>|<ipv4/prefixlen> exact-match)
    quagga (add|del) ipv6 access-list <text> (deny|permit) (any|<ipv6/prefixlen>|<ipv6/prefixlen> exact-match)
    qaugga show ip access-list [<1-2699>|<text>]
    qaugga show ipv6 access-list [<text>]
'''

from collections import OrderedDict
from copy import deepcopy
from network_docopt import NetworkDocopt, get_network_docopt_info, sort_for_humans
from nclu import (print_diff,
    files_match,
    chown_chgroup,
    file_contains,
    ifname_is_valid,
    ifname_is_globish,
    ifname_is_glob,
    ifname_expand_glob,
    print_pending_add_del_commands)
from os import listdir
from re import search as re_search
from pprint import pformat, pprint
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
import json
import logging
import logging.handlers
import os
import re
import sys

log = logging.getLogger('netd')

# pim commands are added to __doc_ if /usr/lib/quagga/pimd exits
# This happens via build_doc_string()
pim_doc_string = """# L3 multicast debugs
    quagga (add|del) debug igmp [events|packets|trace]
    quagga (add|del) debug pim [events|packets|trace|zebra]

    # L3 multicast show commands
    quagga show igmp (groups|interface|join|parameters|querier|sources)
    quagga show mroute [count]
    quagga show pim (address|assert|assert-internal|assert-metric|assert-winner-metric|designated-router|hello|interface|join|jp-override-interval|lan-prune-delay|local-membership|neighbor|rpf|secondary|upstream|upstream-join-desired|upstream-rpf)

    # L3 multicast configuration
    quagga (add|del) ip multicast-routing
    quagga add       ip pim rp <ipv4>
    quagga del       ip pim rp [<ipv4>]
    quagga (add|del) interface <interface> ip igmp
    quagga (add|del) interface <interface> ip igmp join <ipv4-mcast-group> <ipv4-source>
    quagga add       interface <interface> ip igmp query-interval <1-1800>
    quagga del       interface <interface> ip igmp query-interval [<1-1800>]
    quagga add       interface <interface> ip igmp query-max-response-time <1-25>
    quagga del       interface <interface> ip igmp query-max-response-time [<1-25>]
    quagga add       interface <interface> ip pim drpriority <1-4294967295>
    quagga del       interface <interface> ip pim drpriority [<1-4294967295>]
    quagga add       interface <interface> ip pim hello <1-180> <3-180>
    quagga del       interface <interface> ip pim hello [<1-180> <3-180>]
    quagga add       interface <interface> ip pim (sm|ssm)
    quagga del       interface <interface> ip pim [sm|ssm]
"""


def build_doc_string():
    global __doc__

    if os.path.exists('/usr/lib/quagga/pimd'):
        __doc__ = __doc__.replace('    # L3 OSPF configuration', pim_doc_string + '\n    # L3 OSPF configuration')

class ConfigWrapper(object):

    def __init__(self, show_linux_command):
        self.working_dir = '/var/run/nclu/quagga/'
        self.baseline_conf = os.path.join(self.working_dir, 'baseline.conf')
        self.pending_conf = os.path.join(self.working_dir, 'pending.conf')
        self.pending_baseline_conf = os.path.join(self.working_dir, 'pending_baseline.conf')
        self.baseline_daemons = os.path.join(self.working_dir, 'baseline.daemons')
        self.pending_daemons = os.path.join(self.working_dir, 'pending.daemons')
        self.show_linux_command = show_linux_command

        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir, mode=0600)

    def init_guts(self, print_missing_file_errors=True):

        if not os.path.exists('/etc/quagga/daemons'):
            if print_missing_file_errors:
                print "ERROR: /etc/quagga/daemons does not exist"
            return False

        if os.path.getsize('/etc/quagga/daemons') == 0:
            if print_missing_file_errors:
                print "ERROR: /etc/quagga/daemons is empty"
            return False

        if not os.path.isfile('/etc/quagga/Quagga.conf'):
            if print_missing_file_errors:
                print """ERROR: /etc/quagga/Quagga.conf does not exist

net only supports quagga when it uses a single config file /etc/quagga/Quagga.conf
To configure quagga to use /etc/quagga/Quagga.conf please run the following:

sudo vtysh -c 'conf t' -c 'service integrated-vtysh-config'
sudo vtysh -c 'wr mem'
"""
            return False

        if not os.path.isfile(self.baseline_daemons):
            copyfile('/etc/quagga/daemons', self.baseline_daemons)
            copyfile(self.baseline_daemons, self.pending_daemons)

        # Use baseline.conf to later determine if Quagga.conf has been modified
        if not os.path.isfile(self.baseline_conf):
            copyfile('/etc/quagga/Quagga.conf', self.baseline_conf)

        # In the beginning our staging file is a "marked" copy of Quagga.conf
        # "marked" means that each context ends with the word "end"
        if not os.path.isfile(self.pending_conf) or not os.path.isfile(self.pending_baseline_conf):
            try:
                output = check_output(['/usr/bin/vtysh',
                                       '-m',
                                       '-f',
                                       '/etc/quagga/Quagga.conf'],
                                       stderr=STDOUT).splitlines()
            except CalledProcessError as e:
                print "ERROR marking /etc/quagga/Quagga.conf\n"
                print e.output
                return False
            else:
                with open(self.pending_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

                with open(self.pending_baseline_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

        # Verify quagga's config has not been modified
        if not files_match('/etc/quagga/daemons', self.baseline_daemons):
            print "\nERROR: quagga's /etc/quagga/daemons file has been modified"
            print_pending_add_del_commands()
            return False

        if not files_match('/etc/quagga/Quagga.conf', self.baseline_conf):
            print "\nERROR: quagga's running config has been modified"
            print_pending_add_del_commands()
            return False

        return True

    def look_for_line(self, target_line, exact_match):
        """
        Look for the target_line in pending_baseline_conf and pending_conf
        If we find the line, return it else return None
        """
        with open(self.pending_baseline_conf, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()

                if exact_match:
                    if line == target_line:
                        return line
                else:
                    if line.startswith(target_line):
                        return line

        with open(self.pending_conf, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()

                if exact_match:
                    if line == target_line:
                        return line
                else:
                    if line.startswith(target_line):
                        return line

        return None

    def get_bgp_asn(self):
        router_bgp = self.look_for_line('router bgp', False)

        if router_bgp:
            return int(router_bgp.split()[2])
        else:
            return None

    def neighbor_interface(self, neighbor):
        """
        Return True if 'neighbor' is already configured as an interface neighbor
        """
        target_line = 'neighbor %s interface' % neighbor

        if self.look_for_line(target_line, False):
            return True
        else:
            return False

    def neighbor_remote_as(self, neighbor):
        """
        Return True if 'neighbor' already has a remote-as configured
        """

        # First see if the neighbor has remote-as configured directly on the neighbor
        neighbor_remote_as = 'neighbor %s remote-as' % neighbor

        if self.look_for_line(neighbor_remote_as, False):
            return True

        # If not, see if the peer is a member of a peer-group where the
        # peer-group has remote-as configured
        neighbor_peer_group = 'neighbor %s peer-group' % neighbor
        neighbor_peer_group = self.look_for_line(neighbor_peer_group, True)

        if neighbor_peer_group:
            peer_group = neighbor_peer_group.split()[3]
            peer_group_remote_as = 'neighbor %s remote-as' % peer_group

            if self.look_for_line(peer_group_remote_as, False):
                return True

        return False

    def edit_pending_file(self, target_context, target_subcontext, add_line, del_line, exact_match):

        if not self.init_guts():
            return False

        if target_subcontext:
            subcontext_display = "\n %s\n" % target_subcontext
        else:
            subcontext_display = ''

        if add_line:
            action = 'Adding'
            add_line = add_line.rstrip()
            line_display = "\n %s\n" % add_line
            found_del_line = True

        if del_line:
            action = 'Deleting'
            del_line = del_line.strip()
            line_display = "\n %s\n" % del_line
            found_del_line = False

        log.debug("%s '%s' %s %s" % (action, target_context, subcontext_display, line_display))

        # Stage an update to the daemons file.  Most of the time we won't need
        # to modify this file.
        daemons = OrderedDict()
        rewrite_daemons = False

        with open(self.pending_daemons, 'r') as fh:
            for line in fh.readlines():
                line = line.rstrip()

                if not line or line.startswith('#'):
                    continue

                if '=' in line:
                    (daemon, status) = line.split('=')
                    daemon = daemon.strip()
                    status = status.strip()
                    daemons[daemon] = status

        # zebra must always be enabled
        if daemons.get('zebra') != 'yes':
            daemons['zebra'] = 'yes'
            rewrite_daemons = True

        # If this is multicast related command then pimd must be enabled
        if (add_line and ('igmp' in add_line or 'pim' in add_line or 'multicast' in add_line) and
            os.path.exists('/usr/lib/quagga/pimd')):
            if daemons.get('pimd') != 'yes':
                daemons['pimd'] = 'yes'
                rewrite_daemons = True

        # target_context is None for commands that go in the global
        # context (static routes, debugs, etc)
        if target_context:

            # Enable bgpd if needed
            if target_context.startswith('router bgp'):
                if daemons.get('bgpd') != 'yes':
                    daemons['bgpd'] = 'yes'
                    rewrite_daemons = True

            # Enable ospfd if needed
            elif target_context.startswith('router ospf'):
                if daemons.get('ospfd') != 'yes':
                    daemons['ospfd'] = 'yes'
                    rewrite_daemons = True

        if rewrite_daemons:
            with open(self.pending_daemons, 'w') as fh:
                for daemon in daemons.keys():
                    fh.write("%s=%s\n" % (daemon, daemons[daemon]))

        # Stage config changes to pending_conf
        # r+ opens the file for reading and writing
        with open(self.pending_conf, 'r+') as fh:

            # Read in all lines and the reset the fh pointer back to
            # the start of the file
            lines = fh.readlines()
            fh.seek(0)

            state = 'LIMBO'
            subcontext_lines = []

            # Now write all of the lines back to the file but add the add_line
            # (if there is one) at the end of its context/subcontext.
            for line in lines:
                line = line.rstrip()

                if state == 'LIMBO':
                    # If we are adding lines at the global level (static routes, debugs, etc)
                    # then target_context will be None
                    if target_context:

                        if line == target_context:
                            state = 'IN_CONTEXT'

                        # Will need to revisit this when we add vrf support
                        elif target_context.startswith('router bgp') and line.startswith('router bgp '):
                            state = 'IN_CONTEXT'

                elif state == 'IN_CONTEXT':
                    if target_subcontext and line.strip() == target_subcontext:
                        state = 'IN_SUBCONTEXT'

                    elif line.strip() == 'end':
                        # We just finished writing all of the lines for our context, if
                        # add_line wasn't already in there then write it too
                        if add_line:

                            # If our add_line is CONTEXT there isn't anything
                            # to do because the context already exists
                            if add_line != 'CONTEXT':
                                if add_line in subcontext_lines:
                                    print "'%s' configuration already has '%s'" %\
                                        (target_context, add_line)
                                else:

                                    # Create the subcontext
                                    if target_subcontext:
                                        fh.write(' ' + target_subcontext + '\n')
                                        fh.write('  ' + add_line + '\n')
                                        fh.write(' exit-address-family\n')
                                    else:
                                        fh.write(' ' + add_line + '\n')
                            add_line = None

                        # If we are deleting the entire context we must also delete the 'end'
                        # Note that the 'interface' check here is because in quagga we cannot
                        # do a "no" on an entire interface. Tracking this via CM-11563
                        elif del_line == 'CONTEXT' and not target_context.startswith('interface '):
                            state = 'LIMBO'
                            continue

                        state = 'LIMBO'

                elif state == 'IN_SUBCONTEXT':
                    if line.strip() == 'exit-address-family' or line.strip() == 'end':

                        # We just finished writing all of the lines for our subcontext, if
                        # add_line wasn't already in there then write it too
                        if add_line and add_line != 'CONTEXT':
                            if add_line in subcontext_lines:
                                print "'%s %s' configuration already has '%s'" %\
                                    (target_context, target_subcontext, add_line)
                            else:
                                fh.write('  ' + add_line + '\n')
                        add_line = None
                        state = 'IN_CONTEXT'

                    elif line.strip() == 'end':
                        state = 'LIMBO'

                log.debug("%s: %s" % (state, line))

                if (not target_context or
                    (not target_subcontext and state == 'IN_CONTEXT') or
                    (target_subcontext and state == 'IN_SUBCONTEXT')):

                    # We are deleting the entire context so do not write any of the lines
                    # Note that the 'interface' check here is because in quagga we cannot
                    # do a "no" on an entire interface. Tracking this via CM-11563
                    if del_line == 'CONTEXT' and not line.startswith('interface '):
                        pass

                    # Do we write the current line?
                    elif exact_match and line.strip() == del_line:
                        found_del_line = True

                    elif not exact_match and line.strip().startswith(del_line):
                        found_del_line = True

                    else:
                        # If the user is adding a line to the global context
                        # do not add the same line multiple times
                        if not target_context and line == add_line:
                            print "configuration already has '%s'" % add_line
                        else:
                            fh.write(line + '\n')

                    subcontext_lines.append(line.strip())
                else:
                    fh.write(line + '\n')

            if add_line:
                log.debug("STATE %s, ADD_LINE: %s" % (state, add_line))

                # Note that CONTEXT is a special case.  We hit this if
                # the user is trying to add a context such as "router bgp 100"
                if add_line == 'CONTEXT':
                    if state not in ('IN_CONTEXT', 'IN_SUBCONTEXT'):
                        fh.write('%s\n' % target_context)

                        if target_subcontext:
                            fh.write(' %s\n' % target_subcontext)
                            fh.write(' exit-address-family\n')
                        fh.write('end\n')

                # The context (or sub-context) that we were looking wasn't in the config
                else:

                    # Enter the context...if it is None that means global context
                    if target_context:
                        fh.write('%s\n' % target_context)

                        # We are in the context, enter the sub-context if there is
                        # one. Indent accordingly
                        if target_subcontext:
                            fh.write(' %s\n' % target_subcontext)
                            fh.write('  %s\n' % add_line)
                            fh.write(' exit-address-family\n')
                        else:
                            fh.write(' %s\n' % add_line)
                        fh.write('end\n')
                    else:
                        fh.write('%s\n' % add_line)
                        fh.write('end\n')

            # If we deleted lines we need to truncate the rest of the fh so that we
            # do not end up with the lines at the end printed twice
            fh.truncate()

        if not found_del_line and del_line != 'CONTEXT':
            if target_subcontext:
                print "'%s %s' configuration does not have '%s'" % (target_context, target_subcontext, del_line)
            elif target_context:
                print "'%s' configuration does not have '%s'" % (target_context, del_line)
            else:
                print "configuration does not have '%s'" % del_line

    def bgp_neighbor_sanity(self, context, subcontext, cmd):
        # If we are adding a keyword for a bgp neighbor there are a few sanity
        # checks that we must make
        # - Is the peer defined?
        # - If the peer's name matches an interface is the peer defined as an "interface" peer?
        cmd = cmd.split()

        if context and context.startswith('router bgp') and cmd[0].startswith('neighbor'):
            neighbor = cmd[1]
            keyword = cmd[2]
            interfaces = [x for x in listdir('/sys/class/net/') if x != 'bonding_masters']

            # interface peer
            if neighbor in interfaces:
                if keyword != 'interface' and not self.neighbor_interface(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s interface' % neighbor,
                                           None, True)

            # ipv4/ipv6 peer
            else:
                if keyword not in ('remote-as', 'peer-group') and not self.neighbor_remote_as(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s remote-as external' % neighbor,
                                           None, True)


    def add(self, context, subcontext, cmd):
        self.bgp_neighbor_sanity(context, subcontext, cmd)
        self.edit_pending_file(context, subcontext, cmd, None, True)

    def delete(self, context, subcontext, cmd, exact_match):
        self.edit_pending_file(context, subcontext, None, cmd, exact_match)

    def abort_pending(self):
        for filename in (self.baseline_conf, self.pending_conf, self.pending_baseline_conf,
                         self.baseline_daemons, self.pending_daemons):
            if os.path.isfile(filename):
                os.unlink(filename)

    def commit_pending(self, verbose):
        """
        Make changes persistent by running quagga-reload.py
        """
        files_modified = []
        restart_quagga = False

        if (not os.path.isfile(self.pending_baseline_conf) or
            not os.path.isfile(self.pending_conf) or
            not self.init_guts()):
            return files_modified

        # If there is a delta, /usr/bin/diff will raise CalledProcessError
        try:
            check_output(['/usr/bin/diff', self.baseline_daemons, self.pending_daemons])

            try:
                check_output(['/bin/systemctl', 'is-active', 'quagga'])
            except CalledProcessError as e:
                restart_quagga = True
                print "Starting quagga to enable daemons in /etc/quagga/daemons"

        except CalledProcessError as e:
            copyfile(self.pending_daemons, '/etc/quagga/daemons')
            chown_chgroup('/etc/quagga/daemons', 'quagga', 'quagga')

            # A new daemon was enabled in /etc/quagga/daemons so we must
            # restart quagga so that the new daemon is started. Post CM-11625
            # we can do a start here which will be much less disruptive.
            restart_quagga = True
            print "Restarting quagga to enable new daemons in /etc/quagga/daemons"

        if restart_quagga:

            try:
                output = check_output(['/bin/systemctl', 'reset-failed', 'quagga'])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'restart', 'quagga'])
                if verbose:
                    print output

            except Exception:
                print "ERROR: Failed to restart quagga"

            files_modified.append('/etc/quagga/daemons')

        os.unlink(self.baseline_daemons)
        os.unlink(self.pending_daemons)

        # If there is a delta, /usr/bin/diff will raise CalledProcessError
        # If there is not a delta to apply do not bother calling quagga-reload.py
        try:
            check_output(['/usr/bin/diff', self.pending_baseline_conf, self.pending_conf])
            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            return files_modified
        except CalledProcessError as e:
            pass

        try:
            # The /var/run/nclu/quagga/ directly is locked down (we want to keep it this way)
            # but /var/run/nclu/ is open for read so cp pending_conf to /var/run/nclu/ so it
            # is in a place where quagga-reload.py can read it
            tmp_pending_conf = '/var/run/nclu/quagga_pending.conf'
            copyfile(self.pending_conf, tmp_pending_conf)

            if self.show_linux_command:
                print "Updating /etc/quagga/Quagga.conf and applying changes via quagga-reload.py"

            output = check_output(['/usr/lib/quagga/quagga-reload.py',
                                   '--stdout',
                                   '--reload',
                                   '/var/run/nclu/quagga_pending.conf'], stderr=STDOUT)
            os.unlink(tmp_pending_conf)

            if verbose:
                print output

            output = check_output(['/usr/bin/vtysh', '-c', 'wr'], stderr=STDOUT)
            if verbose:
                print output

            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            files_modified.append('/etc/quagga/Quagga.conf')
        except CalledProcessError as e:
            log.exception(e)
            log.error(e.output)
            print "\nERROR: quagga-reload.py failed, is quagga running? Check 'systemctl status quagga'"
            print "\n%s\n%s\n%s\n" % ('=' * 80, e.output, '=' * 80)

        return files_modified

    def show_pending(self, print_errors=True):

        if not self.init_guts(print_missing_file_errors=False):
            return False

        diff_pending = False

        if print_diff(self.baseline_daemons, self.pending_daemons):
            diff_pending = True

        if print_diff(self.pending_baseline_conf, self.pending_conf):
            diff_pending = True

        return diff_pending

    def show_config(self, user_may_edit):
        print "/etc/quagga/Quagga.conf"
        print "======================="

        if self.show_linux_command:
            print "exec 'cat /etc/quagga/Quagga.conf'\n"

        if os.path.exists('/etc/quagga/Quagga.conf'):
            # If the user has edit privs then do not bother hiding passwords
            if user_may_edit:
                with open('/etc/quagga/Quagga.conf', 'r') as fh:
                    print ''.join(fh.readlines())

            # If they can only do "show" commands then replace any passwords with PASSWORD_REMOVED
            else:
                output = []
                with open('/etc/quagga/Quagga.conf', 'r') as fh:
                    for line in fh.readlines():

                        if 'authentication-key' in line:
                            re_ospf_auth_key = re.search('^(.* ospf authentication-key\s+)\S+(.*)$', line)
                            if re_ospf_auth_key:
                                line = re_ospf_auth_key.group(1) + 'PASSWORD_REMOVED' + re_ospf_auth_key.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)

                        elif 'password' in line and 'password 8 ' not in line:
                            re_password = re.search('^(.*password\s+)\S+(.*)$', line)

                            if re_password:
                                line = re_password.group(1) + 'PASSWORD_REMOVED' + re_password.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)
                        else:
                            output.append(line)
                print ''.join(output)

    def ip_route_show(self, vtysh_args, zebra_output):

        def get_routing_entry_for():
            for line in zebra_output.splitlines():
                if line.startswith('Routing entry for'):
                    return line.strip().split()[3]
            return None

        # show ip route x.x.x.x/24
        # ip route show x.x.x.x/24
        #
        # show ip route x.x.x.x
        # ip route show
        if vtysh_args.startswith('show ip route'):
            re_prefix_prefixlen = re_search('^show ip route (\d+\.\d+\.\d+\.\d+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ip route (\d+\.\d+\.\d+\.\d+)$', vtysh_args)
            use_ipv6 = False

        elif vtysh_args.startswith('show ipv6 route'):
            re_prefix_prefixlen = re_search('^show ipv6 route ([0-9a-fA-F]\S+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ipv6 route ([0-9a-fA-F]\S+:\S+)$', vtysh_args)
            use_ipv6 = True

        else:
            re_prefix_prefixlen = None
            re_prefix = None
            use_ipv6 = False

        if re_prefix_prefixlen:
            ip = re_prefix_prefixlen.group(1)

        elif re_prefix:
            ip = re_prefix.group(1)
            ip = get_routing_entry_for()
        else:
            ip = None

        if ip:
            header = "FIB entry for %s" % vtysh_args.split()[-1]
            print "%s\n%s" % (header, '=' * len(header))

            if self.show_linux_command:
                if use_ipv6:
                    print "exec: '/sbin/ip -6 route show %s'" % ip
                else:
                    print "exec: '/sbin/ip route show %s'" % ip

            try:
                if use_ipv6:
                    print check_output(['/sbin/ip', '-6', 'route', 'show', ip])
                else:
                    print check_output(['/sbin/ip', 'route', 'show', ip])
            except Exception:
                pass

    def del_interfaces(self, ifname):
        ifnames = []

        if ifname_is_glob(str(ifname)):
            for x in ifname_expand_glob(str(ifname)):
                ifnames.append(x)

        elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
            ifnames = [ifname, ]

        for ifname in ifnames:
            context = 'interface %s' % ifname
            subcontext = None
            cmd = 'CONTEXT'
            self.delete(context, subcontext, cmd, True)

    def exec_vtysh_show_or_clear(self, args, vtysh_args):
        # In quagga, bgp allows the user to do "show ip bgp" or "show bgp"
        # but ospf only allows "show ip ospf".  This is a bit awkward for
        # the user if they are working with both so we made "show [ip] ospf".
        # If the user entered "show ospf" we need to change that to "show ip ospf"
        vtysh_args = vtysh_args.replace('show ospf', 'show ip ospf')
        vtysh_args = vtysh_args.replace('show igmp', 'show ip igmp')
        vtysh_args = vtysh_args.replace('show mroute', 'show ip mroute')
        vtysh_args = vtysh_args.replace('show pim', 'show ip pim')

        vtysh_args = vtysh_args.replace('clear ospf', 'clear ip ospf')
        vtysh_args = vtysh_args.replace('clear bgp', 'clear ip bgp')
        vtysh_args_list = []

        if vtysh_args.startswith('show bgp'):
            # Have this dump both IPv4 and IPv6 tables
            if vtysh_args == 'show bgp':
                vtysh_args_list.append('show bgp ipv4 unicast')
                vtysh_args_list.append('show bgp ipv6 unicast')

            # Have this dump both IPv4 and IPv6 summary
            elif vtysh_args == 'show bgp summary':
                vtysh_args_list.append('show bgp ipv4 unicast summary')
                vtysh_args_list.append('show bgp ipv6 unicast summary')

            elif vtysh_args == 'show bgp update-groups':
                vtysh_args_list.append('show bgp ipv4 unicast update-groups')
                vtysh_args_list.append('show bgp ipv6 unicast update-groups')

            # Nothing special for these, just run as they are
            elif (vtysh_args.startswith('show bgp ipv4 unicast') or
                  vtysh_args.startswith('show bgp ipv6 unicast') or
                  vtysh_args.startswith('show bgp neighbor')):
                vtysh_args_list.append(vtysh_args)

            # If we are here then we are running this command:
            # quagga show bgp (<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
            #
            # So we must look to see if we are dealing with an IPv4 prefix or an IPv6 prefix
            # and add 'ipv4 unicast' or 'ipv6 unicast' to the command
            else:
                if args.get('<ipv4>') or args.get('<ipv4/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv4 unicast')

                elif args.get('<ipv6>') or args.get('<ipv6/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv6 unicast')

                vtysh_args_list.append(vtysh_args)

        # quagga show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|ospf|static|summary]
        # quagga show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes
        elif vtysh_args.startswith('show route') and not vtysh_args.startswith('show route-map'):

            if args.get('<ipv6>') or args.get('<ipv6/prefixlen>'):
                vtysh_args = vtysh_args.replace('show route', 'show ipv6 route')

            else:
                vtysh_args = vtysh_args.replace('show route', 'show ip route')

            # When ip_route_show() is called we will print the kernel entry for this route
            # so print a little header here so the user can tell which is which
            header = "RIB entry for %s" % vtysh_args.split()[-1]
            print "%s\n%s" % (header, '=' * len(header))
            vtysh_args_list.append(vtysh_args)

        elif 'clear ip bgp all' in vtysh_args:
            vtysh_args_list.append(vtysh_args.replace('clear ip bgp all', 'clear ip bgp *'))

        else:
            vtysh_args_list.append(vtysh_args)

        print_header = True if len(vtysh_args_list) > 1 else False

        for vtysh_args in vtysh_args_list:
            cmd = ['/usr/bin/vtysh', '-c', vtysh_args]

            if self.show_linux_command and 'json' not in vtysh_args:
                print "exec 'vtysh -c \"%s\"'\n" % vtysh_args

            if print_header:
                print "\n%s\n%s" % (vtysh_args, '=' * len(vtysh_args))

            try:
                output = check_output(cmd)
            except Exception as e:
                log.error("'%s' returned '%s' with error:\n%s\n" % (' '.join(cmd), e.output.rstrip(), e))
                print e.output
                output = None

            if output:
                if 'json' in vtysh_args:
                    # If the command returned json the pprint it, else use regular print
                    try:
                        output = json.loads(output)
                        pprint(output)
                    except ValueError:
                        print output
                else:
                    print output

        # For "show route" also show the kernel route
        if vtysh_args.startswith('show ip route') or vtysh_args.startswith('show ipv6 route'):
            self.ip_route_show(vtysh_args, output)


    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user
        """

        if not self.init_guts():
            return False

        args = cli.args

        bgp_bgp = ['always-compare-med',
                   'bestpath',
                   'client-to-client reflection',
                   'cluster-id',
                   'confederation peers',
                   'default ipv4-unicast',
                   'default local-preference',
                   'default show-hostname',
                   'default subgroup-pkt-queue-max',
                   'deterministic-med',
                   'disable-ebgp-connected-route-check',
                   'enforce-first-as',
                   'fast-external-failover',
                   'graceful-restart',
                   'listen',
                   'log-neighbor-changes',
                   'max-med',
                   'network import-check',
                   'route-map delay-timer',
                   'route-reflector allow-outbound-policy',
                   'router-id']
        ospf_ospf = ['router-id', ]
        ifnames = []

        # If <interface> is a glob then populate ifnames with all of the
        # interface names produced by that glob else populate ifnames with <interface>
        if args.get('<interface>'):
            ifname = args.get('<interface>')

            if ifname_is_glob(str(ifname)):
                for x in ifname_expand_glob(str(ifname)):
                    ifnames.append(x)

            elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
                ifnames = [ifname, ]

            else:
                print "ERROR: a valid interface name or glob is needed"
                return

        if args.get('show') or args.get('clear'):
            cmd = []

            for node in cli.match[1:]:
                if node.value:
                    if node.value.text == '<interface>':
                        cmd.append(node.value.text)
                    elif node.value.text.startswith('<'):
                        cmd.append(node.value.value)
                    else:
                        cmd.append(node.value.text)

                # Default afi to 'ipv4' and safi to 'unicast'
                else:
                    if str(node) == '[ipv4|ipv6]':
                        cmd.append('ipv4')
                    elif str(node) == '[unicast|multicast]':
                        cmd.append('unicast')
                    elif str(node) == '[ipv4 unicast|ipv6 unicast]':
                        cmd.append('ipv4')
                        cmd.append('unicast')

            vtysh_args = ' '.join(map(str, cmd))

            # Support show commands with globs of interfaces
            # net show ospf interface swp3-4
            if ifnames:
                for ifname in ifnames:
                    vtysh_args_for_ifname = vtysh_args.replace('<interface>', ifname)
                    self.exec_vtysh_show_or_clear(args, vtysh_args_for_ifname)
            else:
                self.exec_vtysh_show_or_clear(args, vtysh_args)

        elif args.get('add') or args.get('del'):
            contexts = []
            subcontext = None

            # (add|del) will be the node at #0
            # entry #1 will be the bgp, ospf, interface, etc node
            add_del_node = cli.match[1]
            context_node = cli.match[2]

            # Set node_cmd_start to 3 so we don't look at "(add|del) bgp" when
            # building the actual quagga cmd string
            node_cmd_start = 3

            # If the user entered "bgp autonomous-system 100"
            # We must translate that to "router bgp 100"
            if context_node.value.text == 'bgp' and cli.match[3].value:
                keyword = cli.match[3].value.text

                if keyword == 'autonomous-system':
                    cli.match[2].value.value = 'router'
                    cli.match[2].value.value_raw = 'router'
                    cli.match[3].value.value = 'bgp'
                    cli.match[3].value.value_raw = 'bgp'

            cmd_is_CONTEXT = False

            # Find the context and subcontext
            if context_node.value.value == 'router':
                protocol = cli.match[3].value.value

                if len(cli.match) > 4:
                    asn = cli.match[4].value.value
                    contexts.append("router %s %d" % (protocol, asn))
                else:
                    contexts.append("router %s" % protocol)

                cmd_is_CONTEXT = True

            elif context_node.value.value == 'bgp':
                bgp_asn = self.get_bgp_asn()

                if not bgp_asn:
                    print "ERROR: You must first enable BGP via 'net add bgp autonomous-system NUMBER'"
                    return

                contexts.append('router bgp %d' % bgp_asn)

                # find the subcontext for bgp
                if str(cli.match[3]) == '[ipv4|ipv6]':
                    ipv4_ipv6_node = cli.match[3]
                    uni_multi_node = cli.match[4]

                    if ipv4_ipv6_node.value and ipv4_ipv6_node.value.text:
                        afi = ipv4_ipv6_node.value.text
                    else:
                        afi = 'ipv4'

                    if uni_multi_node.value and uni_multi_node.value.text:
                        safi = uni_multi_node.value.text
                    else:
                        safi = 'unicast'

                    subcontext = "address-family %s %s" % (afi, safi)
                    node_cmd_start += 2

                elif str(cli.match[3]) == '[ipv4]' and str(cli.match[4]) == '[unicast]':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[3]) == '[ipv6]' and str(cli.match[4]) == '[unicast]':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                elif str(cli.match[3]) == 'ipv4' and str(cli.match[4]) == 'unicast':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[3]) == 'ipv6' and str(cli.match[4]) == 'unicast':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                # The "ipv4 unicast" is optional (just to make v4 easier to configure)
                # for any AFI/SAFI specific command. If they left off the "ipv4 unicast"
                # add it for them.
                if subcontext is None:
                    for x in ('aggregate-address',
                              'network',
                              'redistribute',
                              'maximum-paths',
                              'activate',
                              'allowas-in',
                              'default-originate',
                              'filter-list',
                              'next-hop-self',
                              'prefix-list',
                              'remove-private-AS',
                              'route-map',
                              'route-reflector-client',
                              'send-community'):
                        if x in cli.argv_expanded:
                            subcontext = "address-family ipv4 unicast"
                            break

            elif context_node.value.text == 'ospf':
                contexts.append('router ospf')

            elif context_node.value.text == 'interface':
                if ifnames:
                    for ifname in ifnames:
                        contexts.append('interface %s' % ifname)
                else:
                    contexts.append('interface %s' % cli.match[3].value.text)

                node_cmd_start += 1

            elif context_node.value.value == 'route-map':
                contexts.append(' '.join(map(str, cli.argv_expanded[2:6])))
                node_cmd_start += 3

                # 'net add route-map FOO permit 10' is 6 words
                if len(cli.argv_expanded) == 6:
                    cmd_is_CONTEXT = True

            # global config level
            else:
                node_cmd_start = 2
                contexts.append(None)

            if cmd_is_CONTEXT:
                cmd = 'CONTEXT'
            else:
                cmd = []
                for node in cli.match[node_cmd_start:]:
                    if node.value and node.value.text:
                        if node.text == '<interface>':
                            cmd.append('<interface>')
                        else:
                            cmd.append(node.value.value)

                cmd = ' '.join(map(str, cmd))

            # There are about two dozen commands that are in the global bgp
            # context and start with the keyword 'bgp'. 'bgp router-id x.x.x.x'
            # is one example.  We did not want the parser to read:
            #
            #    quagga add bgp bgp router-id x.x.x.x
            #
            # Because the double bgp just looks/feels weird. Prepend the needed
            # bgp keyword for any of these two dozen commands.
            if contexts and contexts[0]:
                if contexts[0].startswith('router bgp'):
                    for bgp_bgp_cmd in bgp_bgp:
                        if cmd.startswith(bgp_bgp_cmd):
                            cmd = 'bgp ' + cmd
                            break

                # Same problem for some ospf commands
                elif contexts[0].startswith('router ospf'):
                    for ospf_ospf_cmd in ospf_ospf:
                        if cmd.startswith(ospf_ospf_cmd):
                            cmd = 'ospf ' + cmd
                            break

            # If the user is doing a del there are several options where we can
            # del based on the keyword and do not require the value.  For example
            # the user can delete the bgp router-id using either of the following:
            #
            #   net del bgp router-id 1.1.1.1
            #   net del bgp router-id
            #
            # exact_match determines if delete() looks for the line 'bgp router-id 1.1.1.1'
            # or just look for a line that begins with 'bgp router-id'
            if str(add_del_node) == '(add|del)' or str(add_del_node) == 'add':
                exact_match = True

            elif str(add_del_node) == 'del':
                last_node = cli.match[-1]

                # If the user specified the optional arg at the end then
                # exact_match must be True
                if not last_node.value:
                    exact_match = False
                else:
                    exact_match = True

            else:
                raise Exception("Invalid add_del_node %s" % add_del_node)

            if ifnames:
                # If the user entered a glob for interfaces those need to be
                # expanded in one of two places. The first scenario is where
                # the interface name is part of the context.  For example:
                #
                #    net add inter swp3-4 ip ospf network point-to-point
                #
                # needs to add this ospf command under contexts 'interface swp3
                # and 'interface swp4'. But if the user did
                #
                #    net add bgp neighbor swp3-4 interface peer-group FOO
                #
                # we need to expand this as separate commands within the 'router bgp' context.
                #
                # So check the len of contexts to figure out which way we
                # should expand the ifnames
                if len(contexts) > 1:
                    for context in contexts:
                        if args.get('add'):
                            self.add(context, subcontext, cmd)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd, exact_match)
                else:
                    context = contexts[0]

                    for ifname in ifnames:
                        cmd_with_ifname = cmd.replace('<interface>', ifname)

                        if args.get('add'):
                            self.add(context, subcontext, cmd_with_ifname)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd_with_ifname, exact_match)
            else:
                context = contexts[0]

                if args.get('add'):
                    self.add(context, subcontext, cmd)
                elif args.get('del'):
                    self.delete(context, subcontext, cmd, exact_match)

        else:
            print "ERROR: no commands matched"


def get_config_lines(target):
    final = []

    for filename in ('/var/run/nclu/quagga/pending.conf', '/etc/quagga/Quagga.conf'):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith(target):
                        final.append(line)

    return final


def get_aspath_access_list():
    final = []

    for line in get_config_lines('ip as-path access-list'):
        # ip as-path access-list LARRY permit 1234
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgp_peergroups():
    final = []
    for line in get_config_lines('neighbor '):
        if line.endswith('peer-group'):
            final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgppeers():
    final = []
    peergroups = get_bgp_peergroups()

    for line in get_config_lines('neighbor '):
        # neighbor FOO remote-as 100
        final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_community_list():
    final = []

    for line in get_config_lines('ip community-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_extcommunity_list():
    final = []

    for line in get_config_lines('ip extcommunity-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v4():
    final = []

    for line in get_config_lines('ip prefix-list'):
        # ip prefix-list BOB seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v6():
    final = []

    for line in get_config_lines('ipv6 prefix-list'):
        # ipv6 prefix-list JASON seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_route_map():
    final = []

    for line in get_config_lines('route-map'):
        # route-map FOO permit 10
        final.append(line.split()[1])

    final = sorted(list(set(final)))
    return final


def match_aspath_access_list(argv_word, use_fuzzy, tab_completing):
    """
    <as-path-access-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_aspath_access_list():
            return True

        if tab_completing:
            for acl in get_aspath_access_list():
                if acl.startswith(argv_word):
                    return True
    return False


def match_bgppeer(argv_word, use_fuzzy, tab_completing):
    """
    <bgppeer>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_bgppeers():
            return True

        if tab_completing:
            for bgppeer in get_bgppeers():
                if bgppeer.startswith(argv_word):
                    return True
    return False


def match_community(argv_word, use_fuzzy, tab_completing):
    """
    <community>
    There is no fuzzy matching for communities
    """
    if re.search('^\d+:\d+$', argv_word):
        return True
    return False


def match_community_list(argv_word, use_fuzzy, tab_completing):
    """
    <community-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_community_list():
            return True

        if tab_completing:
            for commlist in get_community_list():
                if commlist.startswith(argv_word):
                    return True
    return False


def match_extcommunity(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for extended communities
    ASN:nn_or_IP-address:nn
    """
    if re.search('^\d+:\d+$', argv_word):
        return True

    if re.search('^\d+\.\d+\.\d+\.\d+:\d+$', argv_word):
        return True

    return False


def match_extcommunity_list(argv_word, use_fuzzy, tab_completing):
    """
    <extcommunity-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_extcommunity_list():
            return True

        if tab_completing:
            for extcommlist in get_extcommunity_list():
                if extcommlist.startswith(argv_word):
                    return True
    return False


def match_route_map(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for route-maps
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_route_map():
            return True

        if tab_completing:
            for routemap in get_route_map():
                if routemap.startswith(argv_word):
                    return True

    return False


def match_peer_group(argv_word, use_fuzzy, tab_completing):
    if ' ' not in argv_word:
        return True
    return False


def match_prefix_list_v4(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v4():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v4():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def match_prefix_list_v6(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v6():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v6():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def tab_aspath_access_list(args, ended_with_space, last_argv):
    final = []
    for x in get_aspath_access_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'as-path access-list'))

    if not final:
        final.append(('<as-path-access-list>', 'as-path access-list'))

    return final


def tab_bgppeer(args, ended_with_space, last_argv):
    final = []
    for x in get_bgppeers():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP neighbor or peer-group'))

    if not final:
        final.append(('<bgppeer>', 'BGP neighbor or peer-group'))

    return final


def tab_community_list(args, ended_with_space, last_argv):
    final = []
    for x in get_community_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP community-list'))

    if not final:
        final.append(('<community-list>', 'BGP community-list'))

    return final


def tab_extcommunity_list(args, ended_with_space, last_argv):
    final = []
    for x in get_extcommunity_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP extcommunity-list'))

    if not final:
        final.append(('<extcommunity-list>', 'BGP extcommunity-list'))

    return final


def tab_prefix_list_v4(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v4():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_prefix_list_v6(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v6():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_route_map(args, ended_with_space, last_argv):
    final = []

    # If ended_with_space is True the user did something like "net show route-map <TAB>"
    # so we need to list all route-maps.  If ended_with_space is False they did
    # "net show route-map B<TAB>" so we need to list only the route-maps that begin with 'B'
    for x in get_route_map():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'route-map'))

    if not final:
        final.append(('<route-map>', 'route-map'))

    return final
